# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from os.path import join
import ubelt as ub
import torch  # NOQA
from torch.utils import data as torch_data  # NOQA
import numpy as np
import netharn as nh
import kwimage
import scriptconfig as scfg
from torch.nn import functional as F

import imgaug.augmenters as iaa
import imgaug


class SegmentationConfig(scfg.Config):
    """
    Default configuration for setting up a training session
    """
    default = {
        'nice': scfg.Value('untitled', help='A human readable tag that is "nice" for humans'),
        'workdir': scfg.Path('~/work/sseg', help='Dump all results in your workdir'),

        'workers': scfg.Value(0, help='number of parallel dataloading jobs'),
        'xpu': scfg.Value('argv', help='See netharn.XPU for details. can be cpu/gpu/cuda0/0,1,2,3)'),

        'augmenter': scfg.Value('simple', help='type of training dataset augmentation'),
        'class_weights': scfg.Value('log-median-idf', help='how to weight inbalanced classes'),
        # 'class_weights': scfg.Value(None, help='how to weight inbalanced classes'),

        'datasets': scfg.Value('special:shapes256', help='Either a special key or a coco file'),
        'train_dataset': scfg.Value(None),
        'vali_dataset': scfg.Value(None),
        'test_dataset': scfg.Value(None),

        'arch': scfg.Value('psp', help='Network architecture code'),
        'optim': scfg.Value('adam', help='Weight optimizer. Can be SGD, ADAM, ADAMW, etc..'),

        'backend': scfg.Value('npy', help='fast lookup backnd. may be npy or cog'),
        'input_dims': scfg.Value((224, 224), help='Window size to input to the network'),
        'input_overlap': scfg.Value(0.25, help='amount of overlap when creating a sliding window dataset'),
        'normalize_inputs': scfg.Value(True, help='if True, precompute training mean and std for data whitening'),

        'batch_size': scfg.Value(4, help='number of items per batch'),
        'bstep': scfg.Value(4, help='number of batches before a gradient descent step'),

        'max_epoch': scfg.Value(140, help='Maximum number of epochs'),
        'patience': scfg.Value(140, help='Maximum "bad" validation epochs before early stopping'),

        'lr': scfg.Value(1e-4, help='Base learning rate'),
        'decay':  scfg.Value(1e-5, help='Base weight decay'),

        'focus': scfg.Value(0.0, help='focus for focal loss'),

        'schedule': scfg.Value('step90', help=('Special coercable netharn code. Eg: onecycle50, step50, gamma')),

        'init': scfg.Value('kaiming_normal', help='How to initialized weights. (can be a path to a pretrained model)'),
        'pretrained': scfg.Path(help=('alternative way to specify a path to a pretrained model')),
    }

    def normalize(self):
        if self['pretrained'] in ['null', 'None']:
            self['pretrained'] = None

        if self['pretrained'] is not None:
            self['init'] = 'pretrained'


class SegmentationDataset(torch.utils.data.Dataset):
    """
    Efficient loader for training on a sementic segmentation dataset

    Example:
        >>> # DISABLE_DOCTEST
        >>> #input_dims = (224, 224)
        >>> # xdoctest: +REQUIRES(module:ndsampler)
        >>> import ndsampler
        >>> sampler = ndsampler.CocoSampler.demo('shapes')
        >>> input_dims = (512, 512)
        >>> self = dset = SegmentationDataset(sampler, input_dims)
        >>> output = self[10]
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> cidxs = output['class_idxs']
        >>> colored_labels = self._colorized_labels(cidxs)
        >>> kwplot.figure(doclf=True)
        >>> kwplot.imshow(output['im'])
        >>> kwplot.imshow(colored_labels, alpha=.4)

    Example:
        >>> # xdoctest: +REQUIRES(--interact)
        >>> import xdev
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> indices = list(range(len(self)))
        >>> for index in xdev.InteractiveIter(indices):
        >>>     output = self[index]
        >>>     cidxs = output['class_idxs']
        >>>     colored_labels = self._colorized_labels(cidxs)
        >>>     kwplot.figure(doclf=True)
        >>>     kwplot.imshow(output['im'])
        >>>     kwplot.imshow(colored_labels, alpha=.4)
        >>>     xdev.InteractiveIter.draw()
    """
    def __init__(self, sampler, input_dims=(224, 224), input_overlap=0.5,
                 augmenter=False):
        self.input_dims = None
        self.input_id = None
        self.cid_to_cidx = None
        self.classes = None
        self.sampler = None
        self.subindex = None
        self.gid_to_slider = None
        self._gids = None
        self._sliders = None

        self.sampler = sampler

        self.input_id = self.sampler.dset.hashid

        self.cid_to_cidx = sampler.catgraph.id_to_idx
        self.classes = sampler.catgraph

        # Create a slider for every image
        self._build_sliders(input_dims=input_dims, input_overlap=input_overlap)
        self.augmenter = self._rectify_augmenter(augmenter)

    def _rectify_augmenter(self, augmenter):
        import netharn as nh
        if augmenter is True:
            augmenter = 'simple'

        if not augmenter:
            augmenter = None
        elif augmenter == 'simple':
            augmenter = iaa.Sequential([
                iaa.Crop(percent=(0, .2)),
                iaa.Fliplr(p=.5)
            ])
        elif augmenter == 'complex':
            augmenter = iaa.Sequential([
                iaa.Sometimes(0.2, nh.data.transforms.HSVShift(hue=0.1, sat=1.5, val=1.5)),
                iaa.Crop(percent=(0, .2)),
                iaa.Fliplr(p=.5)
            ])
        else:
            raise KeyError('Unknown augmentation {!r}'.format(self.augment))
        return augmenter

    def _build_sliders(self, input_dims=(224, 224), input_overlap=0.5):
        """
        Use the ndsampler.Sampler and sliders to build a flat index that can
        reach every subregion of every image in the training set.
        """
        import netharn as nh
        gid_to_slider = {}
        for img in self.sampler.dset.imgs.values():
            full_dims = [img['height'], img['width']]
            slider = nh.util.SlidingWindow(full_dims, input_dims,
                                           overlap=input_overlap,
                                           allow_overshoot=True)
            gid_to_slider[img['id']] = slider

        self.input_dims = input_dims
        self.gid_to_slider = gid_to_slider
        self._gids = list(gid_to_slider.keys())
        self._sliders = list(gid_to_slider.values())
        self.subindex = nh.util.FlatIndexer.fromlist(self._sliders)
        return gid_to_slider

    def __len__(self):
        return len(self.subindex)

    def __getitem__(self, index):
        """
        Example:
            >>> # DISABLE_DOCTEST
            >>> self = SegmentationDataset.demo(augment=True)
            >>> output = self[10]
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> plt = kwplot.autoplt()
            >>> colored_labels = self._colorized_labels(output['class_idxs'])
            >>> kwplot.figure(doclf=True)
            >>> kwplot.imshow(output['im'])
            >>> kwplot.imshow(colored_labels, alpha=.4)
        """
        outer, inner = self.subindex.unravel(index)
        gid = self._gids[outer]
        slider = self._sliders[outer]
        slices = slider[inner]

        tr = {'gid': gid, 'slices': slices}
        sample = self.sampler.load_sample(tr, with_annots=['segmentation'])

        imdata = sample['im']
        heatmap = self._sample_to_sseg_heatmap(sample)

        heatmap = heatmap.numpy()
        heatmap.data['class_idx'] = heatmap.data['class_idx'].astype(np.int32)
        cidx_segmap = heatmap.data['class_idx']

        if self.augmenter:
            augdet = self.augmenter.to_deterministic()
            imdata = augdet.augment_image(imdata)
            if hasattr(imgaug, 'SegmentationMapsOnImage'):
                # Oh imgaug, stop breaking.
                cidx_segmap_oi = imgaug.SegmentationMapsOnImage(cidx_segmap, cidx_segmap.shape)
                cidx_segmap_oi = augdet.augment_segmentation_maps(cidx_segmap_oi)
                assert cidx_segmap_oi.arr.shape[2] == 1
                cidx_segmap = cidx_segmap_oi.arr[..., 0]
                cidx_segmap = np.ascontiguousarray(cidx_segmap)
            else:
                cidx_segmap_oi = imgaug.SegmentationMapOnImage(cidx_segmap, cidx_segmap.shape, nb_classes=len(self.classes))
                cidx_segmap_oi = augdet.augment_segmentation_maps([cidx_segmap_oi])[0]
                cidx_segmap = cidx_segmap_oi.arr.argmax(axis=2)

        im_chw = torch.FloatTensor(
            imdata.transpose(2, 0, 1).astype(np.float32) / 255.)

        cidxs = torch.LongTensor(cidx_segmap)
        weight = (1 - (cidxs == 0).float())

        output = {
            'im': im_chw,
            'class_idxs': cidxs,
            'weight': weight,
        }
        return output

    def _sample_to_sseg_heatmap(self, sample):
        imdata = sample['im']
        annots = sample['annots']
        aids = annots['aids']
        cids = annots['cids']
        boxes = annots['rel_boxes']

        # Clip boxes to the image boundary
        input_dims = imdata.shape[0:2]
        boxes = boxes.clip(0, 0, input_dims[1], input_dims[0])

        class_idxs = np.array([self.cid_to_cidx[cid] for cid in cids])
        segmentations = annots['rel_ssegs']

        raw_dets = kwimage.Detections(
            aids=aids,
            boxes=boxes,
            class_idxs=class_idxs,
            segmentations=segmentations,
            classes=self.classes,
            datakeys=['aids'],
        )

        keep = []
        for i, s in enumerate(raw_dets.data['segmentations']):
            # TODO: clip polygons
            m = s.to_mask(input_dims)
            if m.area > 0:
                keep.append(i)
        dets = raw_dets.take(keep)

        heatmap = dets.rasterize(
            bg_size=(1, 1), input_dims=input_dims,
            soften=0,
            exclude=['diameter', 'class_probs', 'offset']
        )

        try:
            # TODO: THIS MAY NOT BE THE CORRECT TRANSFORM
            input_shape = (1, 3,) + input_dims
            output_dims = getattr(self, '_output_dims', None)
            if output_dims is None:
                output_shape = self.raw_model.output_shape_for(input_shape)
                output_dims = self.output_dims = output_shape[2:]
            sf = np.array(output_dims) / np.array(input_dims)
            heatmap = heatmap.scale(sf, output_dims=output_dims, interpolation='nearest')
        except Exception:
            pass

        return heatmap

    def _colorized_labels(self, cidxs):
        self.cx_to_color = np.array([
            self.sampler.dset.name_to_cat[self.classes[cx]]['color']
            for cx in range(len(self.cid_to_cidx))
        ])
        colorized = self.cx_to_color[cidxs]
        return colorized

    @classmethod
    def demo(cls, **kwargs):
        # from grab_camvid import grab_coco_camvid
        # dset = grab_coco_camvid()
        import ndsampler
        sampler = ndsampler.CocoSampler.demo('shapes', workdir=None, backend='npy')
        self = cls(sampler, **kwargs)
        return self


class SegmentationHarn(nh.FitHarn):
    """
    Custom harness to address a basic semantic segmentation problem
    """

    def after_initialize(harn, **kw):
        harn.draw_timer = ub.Timer().tic()

        # hack:
        for k, v in harn.datasets.items():
            v.raw_model = harn.xpu.raw(harn.model)

    def prepare_batch(harn, raw_batch):
        """
        Move a batch onto the XPU
        """
        batch = harn.xpu.move(raw_batch)
        return batch

    def run_batch(harn, batch):
        """
        How to compute a forward pass through the network and compute loss

        Example:
            >>> # xdoctest: +REQUIRES(--slow)
            >>> kw = {'workers': 0, 'xpu': 'cpu', 'batch_size': 2}
            >>> harn = setup_harn(cmdline=False, **kw).initialize()
            >>> batch = harn._demo_batch(tag='train')
            >>> outputs, loss_parts = harn.run_batch(batch)
        """
        im = batch['im']
        class_idxs = batch['class_idxs']

        # You'll almost always have to do something custom to get the batch
        # into the network. Define that here.
        outputs = harn.model(im)
        if not isinstance(outputs, dict):
            outputs = {
                'class_energy': outputs,
            }
        class_energy = outputs['class_energy']

        class_probs = F.softmax(class_energy, dim=1)

        # Heirarchical softmax seems to have a memory leak
        # class_probs = harn.classes.hierarchical_softmax(class_energy, dim=1)

        outputs['class_probs'] = class_probs
        outputs['class_idxs'] = class_idxs

        # You'll almost always have to do something custom to compute loss.
        # Define that here.
        harn.criterion.reduction = 'none'
        pixel_loss = harn.criterion(class_energy, class_idxs)

        pixel_weight = batch['weight']
        clf_loss = (pixel_weight * pixel_loss).sum() / pixel_weight.sum()

        # Return info for netharn to track
        loss_parts = {}
        loss_parts['clf'] = clf_loss

        return outputs, loss_parts

    def on_batch(harn, batch, outputs, loss):
        batch_metrics = {}
        with torch.no_grad():
            # Track accuracy of each batch
            clf_true = outputs['class_idxs'].data.cpu().numpy().astype(np.uint8)
            clf_prob = outputs['class_probs'].data.cpu().numpy().astype(np.float32)
            clf_pred = clf_prob.argmax(axis=1)
            batch_metrics['clf_accuracy'] = (clf_pred == clf_true)[clf_true != 0].mean()

            # Draw first X batches or once every Y minutes
            do_draw = (harn.batch_index <= 8)
            do_draw |= (harn.draw_timer.toc() > 60 * 3)
            if do_draw:
                harn.draw_timer.tic()
                toshow = harn._draw_batch_preds(batch, outputs)
                dpath = ub.ensuredir((harn.train_dpath, 'monitor', harn.current_tag, 'batch'))
                fpath = join(dpath, 'batch{}_epoch{}.jpg'.format(harn.batch_index, harn.epoch))
                kwimage.imwrite(fpath, toshow)
        return batch_metrics

    def _draw_batch_preds(harn, batch, outputs, lim=16):
        """
        Example:
            >>> # xdoctest: +REQUIRES(--slow)
            >>> kw = {'workers': 0, 'xpu': 'cpu', 'batch_size': 8}
            >>> harn = setup_harn(cmdline=False, **kw).initialize()
            >>> batch = harn._demo_batch(tag='train')
            >>> outputs, loss_parts = harn.run_batch(batch)
            >>> toshow = harn._draw_batch_preds(batch, outputs)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.imshow(toshow)
        """
        im = batch['im'].data.cpu().numpy()
        class_true = batch['class_idxs'].data.cpu().numpy()
        class_pred = outputs['class_probs'].data.cpu().numpy().argmax(axis=1)

        batch_imgs = []

        for bx in range(min(len(class_true), lim)):

            orig_img = im[bx].transpose(1, 2, 0)

            import cv2
            out_size = class_pred[bx].shape[::-1]

            orig_img = cv2.resize(orig_img, tuple(map(int, out_size)))
            orig_img = kwimage.ensure_alpha_channel(orig_img)

            pred_heatmap = kwimage.Heatmap(
                class_idx=class_pred[bx],
                classes=harn.classes
            )
            true_heatmap = kwimage.Heatmap(
                class_idx=class_true[bx],
                classes=harn.classes
            )

            # TODO: scale up to original image size

            pred_img = pred_heatmap.draw_on(orig_img, channel='idx', with_alpha=.5)
            true_img = true_heatmap.draw_on(orig_img, channel='idx', with_alpha=.5)

            true_img = kwimage.ensure_uint255(true_img)
            pred_img = kwimage.ensure_uint255(pred_img)

            true_img = kwimage.draw_text_on_image(
                true_img, 'true', org=(0, 0), valign='top', color='blue')

            pred_img = kwimage.draw_text_on_image(
                pred_img, 'pred', org=(0, 0), valign='top', color='blue')

            item_img = kwimage.stack_images([pred_img, true_img], axis=1)
            batch_imgs.append(item_img)

        toshow = kwimage.stack_images_grid(batch_imgs, chunksize=2, overlap=-32)
        return toshow

    def on_complete(harn):
        """
        Ignore:
            >>> kw = {'workers': 0, 'xpu': 'auto', 'batch_size': 8}
            >>> harn = setup_harn(cmdline=False, **kw).initialize()
            >>> harn.datasets['test']
        """
        deployed = harn.raw_model
        out_dpath = ub.ensuredir((harn.train_dpath, 'monitor/test/'))
        # test_dset = harn.datasets['test']
        # sampler = test_dset.sampler

        eval_config = {
            'deployed': deployed,
            'xpu': harn.xpu,
            'out_dpath': out_dpath,
            'do_draw': True,
        }
        print('todo: evalute eval_config = {!r}'.format(eval_config))


class SegmentationModel(nh.layers.Sequential):
    """
    Dummy wrapper around the real model but with input norm
    """
    def __init__(self, arch, classes, in_channels=3, input_stats=None):
        super(SegmentationModel, self).__init__()
        import ndsampler
        if input_stats is None:
            input_stats = {}
        input_norm = nh.layers.InputNorm(**input_stats)

        classes = ndsampler.CategoryTree.coerce(classes)
        self.classes = classes
        if arch == 'unet':
            # Note: UNet can get through 256x256 images at a rate of ~17Hz with
            # batch_size=8. This is pretty slow and can likely be improved by fixing
            # some of the weird padding / mirror stuff I have to do in unet to get
            # output_dims = input_dims.
            from netharn.models.unet import UNet
            model_ = (UNet, {
                'classes': classes,
                'in_channels': in_channels,
            })
        elif arch == 'segnet':
            from netharn.models.segnet import Segnet
            model_ = (Segnet, {
                'in_channels': in_channels,
            })
        elif arch == 'psp':
            from netharn.models.psp import PSPNet_Resnet50_8s
            model_ = (PSPNet_Resnet50_8s, {
                'classes': classes,
                'in_channels': in_channels,
            })
        elif arch == 'deeplab_v3':
            from netharn.models.deeplab_v3 import DeepLabV3
            model_ = (DeepLabV3, {
                'classes': classes,
                'in_channels': in_channels,
            })
        else:
            raise KeyError(arch)

        self.add_module('input_norm', input_norm)
        self.add_module('model', model_[0](**model_[1]))


def _cached_class_frequency(dset, workers=0):
    import ubelt as ub
    import copy
    # Copy the dataset so we can muck with it
    dset_copy = copy.copy(dset)

    dset_copy._build_sliders(input_overlap=0)
    dset_copy.augmenter = None

    cfgstr = '_'.join([dset_copy.sampler.dset.hashid, 'v1'])
    cacher = ub.Cacher('class_freq', cfgstr=cfgstr)
    total_freq = cacher.tryload()
    if total_freq is None:

        total_freq = np.zeros(len(dset_copy.classes), dtype=np.int64)
        if True:
            loader = torch_data.DataLoader(dset_copy, batch_size=16,
                                           num_workers=workers, shuffle=False,
                                           pin_memory=True)

            prog = ub.ProgIter(loader, desc='computing (par) class freq')
            for batch in prog:
                class_idxs = batch['class_idxs'].data.numpy()
                item_freq = np.histogram(class_idxs, bins=len(dset_copy.classes))[0]
                total_freq += item_freq
        else:
            prog = ub.ProgIter(range(len(dset_copy)), desc='computing (ser) class freq')
            for index in prog:
                item = dset_copy[index]
                class_idxs = item['class_idxs'].data.numpy()
                item_freq = np.histogram(class_idxs, bins=len(dset_copy.classes))[0]
                total_freq += item_freq
        cacher.save(total_freq)
    return total_freq


def _precompute_class_weights(dset, workers=0, mode='median-idf'):
    """
    Example:
        >>> # xdoctest: +REQUIRES(--slow)
        >>> harn = setup_harn(0, workers=0, xpu='cpu').initialize()
        >>> dset = harn.datasets['train']
    """

    assert mode in ['median-idf', 'log-median-idf']

    total_freq = _cached_class_frequency(dset, workers=workers)

    def logb(arr, base):
        if base == 'e':
            return np.log(arr)
        elif base == 2:
            return np.log2(arr)
        elif base == 10:
            return np.log10(arr)
        else:
            out = np.log(arr)
            out /= np.log(base)
            return out

    _min, _max = np.percentile(total_freq, [5, 95])
    is_valid = (_min <= total_freq) & (total_freq <= _max)
    if np.any(is_valid):
        middle_value = np.median(total_freq[is_valid])
    else:
        middle_value = np.median(total_freq)

    # variant of median-inverse-frequency
    nonzero_freq = total_freq[total_freq != 0]
    if len(nonzero_freq):
        total_freq[total_freq == 0] = nonzero_freq.min() / 2

    if mode == 'median-idf':
        weights = (middle_value / total_freq)
        weights[~np.isfinite(weights)] = 1.0
    elif mode == 'log-median-idf':
        weights = (middle_value / total_freq)
        weights[~np.isfinite(weights)] = 1.0
        base = 2
        base = np.exp(1)
        weights = logb(weights + (base - 1), base)
        weights = np.maximum(weights, .1)
        weights = np.minimum(weights, 10)
    else:
        raise KeyError('mode = {!r}'.format(mode))

    weights = np.round(weights, 2)
    cname_to_weight = ub.dzip(dset.classes, weights)
    print('weights: ' + ub.repr2(cname_to_weight))
    return weights


def setup_harn(cmdline=True, **kw):
    """
    CommandLine:
        xdoctest -m netharn.examples.segmentation setup_harn

    Example:
        >>> # xdoctest: +REQUIRES(--slow)
        >>> kw = {'workers': 0, 'xpu': 'cpu', 'batch_size': 2}
        >>> cmdline = False
        >>> # Just sets up the harness, does not do any heavy lifting
        >>> harn = setup_harn(cmdline=cmdline, **kw)
        >>> #
        >>> harn.initialize()
        >>> #
        >>> batch = harn._demo_batch(tag='train')
        >>> epoch_metrics = harn._demo_epoch(tag='vali', max_iter=2)
    """
    import sys
    import ndsampler
    import kwarray
    # kwarray.seed_global(2108744082)

    config = SegmentationConfig(default=kw)
    config.load(cmdline=cmdline)
    nh.configure_hacks(config)  # fix opencv bugs

    coco_datasets = nh.api.Datasets.coerce(config)
    print('coco_datasets = {}'.format(ub.repr2(coco_datasets)))
    for tag, dset in coco_datasets.items():
        dset._build_hashid(hash_pixels=False)

    workdir = ub.ensuredir(ub.expandpath(config['workdir']))
    samplers = {
        tag: ndsampler.CocoSampler(dset, workdir=workdir, backend=config['backend'])
        for tag, dset in coco_datasets.items()
    }

    for tag, sampler in ub.ProgIter(list(samplers.items()), desc='prepare frames'):
        try:
            sampler.frames.prepare(workers=config['workers'])
        except AttributeError:
            pass

    torch_datasets = {
        tag: SegmentationDataset(
            sampler,
            config['input_dims'],
            input_overlap=((tag == 'train') and config['input_overlap']),
            augmenter=((tag == 'train') and config['augmenter']),
        )
        for tag, sampler in samplers.items()
    }
    torch_loaders = {
        tag: torch_data.DataLoader(dset,
                                   batch_size=config['batch_size'],
                                   num_workers=config['workers'],
                                   shuffle=(tag == 'train'),
                                   drop_last=True, pin_memory=True)
        for tag, dset in torch_datasets.items()
    }

    if config['class_weights']:
        mode = config['class_weights']
        dset = torch_datasets['train']
        class_weights = _precompute_class_weights(dset, mode=mode,
                                                  workers=config['workers'])
        class_weights = torch.FloatTensor(class_weights)
        class_weights[dset.classes.index('background')] = 0
    else:
        class_weights = None

    if config['normalize_inputs']:
        stats_dset = torch_datasets['train']
        stats_idxs = kwarray.shuffle(np.arange(len(stats_dset)), rng=0)[0:min(1000, len(stats_dset))]
        stats_subset = torch.utils.data.Subset(stats_dset, stats_idxs)
        cacher = ub.Cacher('dset_mean', cfgstr=stats_dset.input_id + 'v3')
        input_stats = cacher.tryload()
        if input_stats is None:
            loader = torch.utils.data.DataLoader(
                stats_subset, num_workers=config['workers'],
                shuffle=True, batch_size=config['batch_size'])
            running = nh.util.RunningStats()
            for batch in ub.ProgIter(loader, desc='estimate mean/std'):
                try:
                    running.update(batch['im'].numpy())
                except ValueError:  # final batch broadcast error
                    pass
            input_stats = {
                'std': running.simple(axis=None)['mean'].round(3),
                'mean': running.simple(axis=None)['std'].round(3),
            }
            cacher.save(input_stats)
    else:
        input_stats = {}

    print('input_stats = {!r}'.format(input_stats))

    # TODO: infer numbr of channels
    model_ = (SegmentationModel, {
        'arch': config['arch'],
        'input_stats': input_stats,
        'classes': torch_datasets['train'].classes.__json__(),
        'in_channels': 3,
    })

    initializer_ = nh.Initializer.coerce(config)
    # if config['init'] == 'cls':
    #     initializer_ = model_[0]._initializer_cls()

    # Create hyperparameters
    hyper = nh.HyperParams(
        nice=config['nice'],
        workdir=config['workdir'],
        xpu=nh.XPU.coerce(config['xpu']),

        datasets=torch_datasets,
        loaders=torch_loaders,

        model=model_,
        initializer=initializer_,

        scheduler=nh.Scheduler.coerce(config),
        optimizer=nh.Optimizer.coerce(config),
        dynamics=nh.Dynamics.coerce(config),

        criterion=(nh.criterions.FocalLoss, {
            'focus': config['focus'],
            'weight': class_weights,
            # 'reduction': 'none',
        }),

        monitor=(nh.Monitor, {
            'minimize': ['loss'],
            'patience': config['patience'],
            'max_epoch': config['max_epoch'],
            'smoothing': .6,
        }),

        other={
            'batch_size': config['batch_size'],
        },
        extra={
            'argv': sys.argv,
            'config': ub.repr2(config.asdict()),
        }
    )

    # Create harness
    harn = SegmentationHarn(hyper=hyper)
    harn.classes = torch_datasets['train'].classes
    harn.preferences.update({
        'num_keep': 2,
        'keyboard_debug': True,
        # 'export_modules': ['netharn'],
    })
    harn.intervals.update({
        'vali': 1,
        'test': 10,
    })
    harn.script_config = config
    return harn


def main():
    harn = setup_harn()
    harn.run()


if __name__ == '__main__':
    """
    CommandLine:

        conda install gdal

        python -m netharn.examples.segmentation \
                --nice=shapes_demo --datasets=shapes32 \
                --workers=0 --xpu=cpu

        # You can use MS-COCO files to learn to segment your own data To
        # demonstrate grab the CamVid dataset (the following script also
        # transforms camvid into the MS-COCO format)

        python -m netharn.data.grab_camvid  # Download MS-COCO files

        python -m netharn.examples.segmentation --workers=4 --xpu=0 --nice=camvid_deeplab \
            --train_dataset=$HOME/.cache/netharn/camvid/camvid-master/camvid-train.mscoco.json \
            --vali_dataset=$HOME/.cache/netharn/camvid/camvid-master/camvid-train.mscoco.json \
            --schedule=step-90-120 --arch=deeplab --batch_size=8 --lr=1e-5 --input_dims=224,224 --optim=sgd --bstep=8

        python -m netharn.examples.segmentation --workers=4 --xpu=auto --nice=camvid_psp_wip \
            --train_dataset=$HOME/.cache/netharn/camvid/camvid-master/camvid-train.mscoco.json \
            --vali_dataset=$HOME/.cache/netharn/camvid/camvid-master/camvid-train.mscoco.json \
            --schedule=step-90-120 --arch=psp --batch_size=6 --lr=1e-3 --input_dims=512,512 --optim=sgd --bstep=1

        # Note you would need to change the path to a pretrained network
        python -m netharn.examples.segmentation --workers=4 --xpu=auto --nice=camvid_psp_wip_fine \
            --train_dataset=$HOME/.cache/netharn/camvid/camvid-master/camvid-train.mscoco.json \
            --vali_dataset=$HOME/.cache/netharn/camvid/camvid-master/camvid-train.mscoco.json \
            --pretrained=$HOME/work/sseg/fit/runs/camvid_psp_wip/fowjplca/deploy_SegmentationModel_fowjplca_134_CZARGB.zip \
            --schedule=step-90-120 --arch=psp --batch_size=6 --lr=1e-2 --input_dims=512,512 --optim=sgd --bstep=8
    """
    main()



"""
References:
    https://github.com/alykhantejani/initializers
"""
import torch
from netharn import api
from netharn.initializers.functional import apply_initializer


class NoOp(api.Initializer):
    """
    An initializer that does nothing, which is useful when you have initialized
    the weights yourself.

    Example:
        >>> from netharn.models import toynet
        >>> import copy
        >>> self = NoOp()
        >>> model = toynet.ToyNet2d()
        >>> old_state = sum(v.sum() for v in model.state_dict().values())
        >>> self(model)
        >>> new_state = sum(v.sum() for v in model.state_dict().values())
        >>> assert old_state == new_state
        >>> assert self.history() is None
    """
    def forward(self, model):
        return


class Orthogonal(api.Initializer):
    """
    Same as Orthogonal, but uses pytorch implementation

    Example:
        >>> from netharn.models import toynet
        >>> self = Orthogonal()
        >>> model = toynet.ToyNet2d()
        >>> try:
        >>>     self(model)
        >>> except RuntimeError:
        >>>     import pytest
        >>>     pytest.skip('geqrf: Lapack probably not availble')
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """
    def __init__(self, gain=1):
        self.gain = gain

    def forward(self, model):
        try:
            func = torch.nn.init.orthogonal_
        except AttributeError:
            func = torch.nn.init.orthogonal

        apply_initializer(model, func, self.__dict__)


class KaimingUniform(api.Initializer):
    """
    Same as HeUniform, but uses pytorch implementation

    Example:
        >>> from netharn.models import toynet
        >>> self = KaimingUniform()
        >>> model = toynet.ToyNet2d()
        >>> self(model)
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """
    def __init__(self, param=0, mode='fan_in'):
        self.a = param
        self.mode = mode

    def forward(self, model):
        try:
            func = torch.nn.init.kaiming_uniform_
        except AttributeError:
            func = torch.nn.init.kaiming_uniform
        apply_initializer(model, func, self.__dict__)


class KaimingNormal(api.Initializer):
    """
    Same as HeNormal, but uses pytorch implementation

    Example:
        >>> from netharn.models import toynet
        >>> self = KaimingNormal()
        >>> model = toynet.ToyNet2d()
        >>> self(model)
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """
    def __init__(self, param=0, mode='fan_in'):
        self.a = param
        self.mode = mode

    def forward(self, model):
        try:
            func = torch.nn.init.kaiming_normal_
        except AttributeError:
            func = torch.nn.init.kaiming_normal
        apply_initializer(model, func, self.__dict__)



# -*- coding: utf-8 -*-
"""
Torch version of hyperparams

TODO:
    [ ] - need to extract relavent params from loaders
    [ ] - need to extract relavent params from datasets
    [ ] - ensure monitor is handled gracefully
    [ ] - prevent non-relevant params from being used in the hash

CommandLine:
    python ~/code/netharn/netharn/hyperparams.py __doc__

Example:
    >>> import netharn as nh
    >>> datasets = {
    >>>     'train': nh.data.ToyData2d(size=3, border=1, n=256, rng=0),
    >>>     'vali': nh.data.ToyData2d(size=3, border=1, n=128, rng=1),
    >>> }
    >>> hyper = nh.HyperParams(**{
    >>>     # --- Data First
    >>>     'datasets'    : datasets,
    >>>     'nice'        : 'demo',
    >>>     'loaders'     : {'batch_size': 64},
    >>>     'xpu'         : nh.XPU.coerce('auto'),
    >>>     # --- Algorithm Second
    >>>     'model'       : (nh.models.ToyNet2d, {}),
    >>>     'optimizer'   : (nh.optimizers.SGD, {
    >>>         'lr': 0.0001
    >>>     }),
    >>>     'criterion'   : (nh.criterions.CrossEntropyLoss, {}),
    >>>     #'criterion'   : (nh.criterions.FocalLoss, {}),
    >>>     'initializer' : (nh.initializers.KaimingNormal, {
    >>>         'param': 0,
    >>>     }),
    >>>     'scheduler'   : (nh.schedulers.ListedLR, {
    >>>         'points': {0: .0001, 2: .01, 5: .015, 6: .005, 9: .001},
    >>>     }),
    >>>     'monitor'     : (nh.Monitor, {
    >>>         'max_epoch': 10
    >>>     }),
    >>> })
    >>> print(ub.repr2(hyper.get_initkw()))
    >>> print(ub.repr2(hyper.hyper_id()))


"""
from __future__ import absolute_import, division, print_function, unicode_literals
import platform
import warnings
from os.path import join
from os.path import normpath
from os.path import sys
import numpy as np
import ubelt as ub
import torch
import six
from netharn import util
from netharn import initializers
from netharn import device
from collections import OrderedDict
# from netharn import criterions
from torch.optim.optimizer import required
import torch.utils.data as torch_data
from netharn.util import util_json
from netharn.util import util_inspect


# backwards compatibility
_ensure_json_serializable = util_json.ensure_json_serializable  # NOQA


try:
    import imgaug
    Augmenter = imgaug.augmenters.meta.Augmenter
except ImportError:
    imgaug = None


def _hash_data(data):
    return ub.hash_data(data, hasher='sha512', base='abc', types=True)


def _rectify_class(arg, kw, lookup=None):
    """
    Helps normalize and serialize hyperparameter inputs.

    Args:
        arg (Tuple[type, dict] | type | object):
            Either a (cls, initkw) tuple, a class, or an instance.
            It is recommended that you don't pass an instance.

        kw (Dict[str, object]):
            augments initkw if arg is in tuple form otherwise becomes initkw

        lookup (func | None):
            transforms arg or arg[0] into the class type

    Returns:
        Dict: containing
            'cls' (type): the type of the object
            'cls_kw' (Dict): the initialization keyword args
            'instance': (object): None or the actual instanciated object

            We will use this cls and cls_kw to construct an instance unless one
            is already specified.

    Example:
        >>> # The ideal case is that we have a cls, initkw tuple
        >>> import netharn as nh
        >>> kw = {'lr': 0.1}
        >>> cls = torch.optim.SGD
        >>> rectified1 = _rectify_class(cls, kw.copy())
        >>> print('rectified1 = {!r}'.format(rectified1))
        >>> # But we can also take an instance of the object, however, you must
        >>> # now make sure to specify the _initkw attribute.
        >>> model = nh.models.ToyNet2d()
        >>> self = cls(model.parameters(), **kw)
        >>> self._initkw = kw
        >>> rectified2 = _rectify_class(self, {})
        >>> print('rectified2 = {!r}'.format(rectified2))
    """
    if lookup is None:
        lookup = ub.identity

    if arg is None:
        cls = None
        cls_kw = {}
        instance = None
    else:
        instance = None

        # Extract the part that identifies the class we care about
        if isinstance(arg, tuple):
            cls_key = arg[0]
            kw2 = arg[1]
        else:
            cls_key = arg
            kw2 = {}

        cls = lookup(cls_key)

        if not isinstance(cls, type):
            # Rectified an instance to an instance
            cls = cls.__class__

        instance = None
        if isinstance(cls_key, cls):
            # We were passed an actual instance of the class. (for shame)
            instance = cls_key

        cls_kw = util_inspect.default_kwargs(cls).copy()

        if instance is not None:
            # Try and introspect the initkw, which is needed for model
            # deployment and proper hyperparam tracking
            for key in cls_kw:
                if hasattr(instance, key):
                    cls_kw[key] = getattr(instance, key)
            if hasattr(instance, '_initkw'):
                # Special attribute that allows safe instance specification and
                # supresses the instance warning.
                cls_kw.update(instance._initkw)
            else:
                import warnings
                warnings.warn(ub.paragraph(  # _initkw warning
                    '''
                    netharn.HyperParams objects are expected to be specified as
                    (type, kw) tuples, but we received a preconstructed
                    instance. This is only ok if you know what you are doing.
                    To disable this warning set the _initkw instance attribute
                    to the correct keyword arguments needed to reconstruct this
                    class. Offending data is arg={!r}, kw={!r}
                    ''').format(arg, kw))

        # Update with explicitly specified information
        cls_kw.update(kw2)
        for key in cls_kw:
            if key in kw:
                cls_kw[key] = kw.pop(key)

    cls_kw = util_json.ensure_json_serializable(cls_kw)
    rectified = {
        'cls': cls,
        'cls_kw': cls_kw,
        'instance': instance,
    }
    return rectified


def _rectify_criterion(arg, kw):
    if arg is None:
        # arg = 'CrossEntropyLoss'
        return _rectify_class(None, kw)

    def _lookup(arg):
        if isinstance(arg, six.string_types):
            options = [
                # criterions.ContrastiveLoss,
                torch.nn.CrossEntropyLoss,
            ]
            cls = {c.__name__: c for c in options}[arg]
        else:
            cls = arg
        return cls

    rectified = _rectify_class(arg, kw, _lookup)
    return rectified


def _rectify_optimizer(arg, kw):
    """
    Create a rectified tuple

    Example:
        >>> # Test using a (cls, kw) tuple and an instance object.
        >>> import netharn as nh
        >>> optim_ = nh.api.Optimizer.coerce({
        >>>     'optim': 'adam', 'lr': 0.1, 'weight_decay': 1e-4})
        >>> cls, kw = optim_
        >>> #
        >>> model = nh.models.ToyNet2d()
        >>> params = dict(model.named_parameters())
        >>> grouped_keys = {}
        >>> grouped_keys['bias'] = [k for k in params.keys() if 'bias' in k]
        >>> grouped_keys['weight'] = [k for k in params.keys() if 'weight' in k]
        >>> named_param_groups = {
        >>>     k: {'params': list(ub.take(params, sorted(v)))}
        >>>     for k, v in grouped_keys.items()
        >>> }
        >>> named_param_groups['bias']['weight_decay'] = 0
        >>> param_groups = list(ub.sorted_keys(named_param_groups).values())
        >>> #
        >>> optim = cls(param_groups, **kw)
        >>> rectified1 = _rectify_optimizer(cls, kw)
        >>> rectified2 = _rectify_optimizer(optim, {})
    """
    if arg is None:
        arg = 'SGD'
        if kw is None:
            kw = {}
        kw = kw.copy()
        if 'lr' not in kw:
            kw['lr'] = .001

    def _lookup(arg):
        if isinstance(arg, six.string_types):
            options = [
                torch.optim.Adam,
                torch.optim.SGD,
            ]
            cls = {c.__name__.lower(): c for c in options}[arg.lower()]
        else:
            cls = arg
        return cls

    rectified = _rectify_class(arg, kw, _lookup)
    kw2 = rectified['cls_kw']

    for k, v in kw2.items():
        if v is required:
            raise ValueError('Must specify {} for {}'.format(k, rectified['cls']))

    return rectified


def _rectify_lr_scheduler(arg, kw):
    if arg is None:
        return _rectify_class(None, kw)

    def _lookup(arg):
        if isinstance(arg, six.string_types):
            options = [
                torch.optim.lr_scheduler.LambdaLR,
                torch.optim.lr_scheduler.StepLR,
                torch.optim.lr_scheduler.MultiStepLR,
                torch.optim.lr_scheduler.ExponentialLR,
                torch.optim.lr_scheduler.ReduceLROnPlateau,
            ]
            cls = {c.__name__: c for c in options}[arg]
        else:
            cls = arg
        return cls

    rectified = _rectify_class(arg, kw, _lookup)
    return rectified


def _rectify_initializer(arg, kw):
    if arg is None:
        arg = 'NoOp'
        # arg = 'CrossEntropyLoss'
        # return None, None

    def _lookup(arg):
        if isinstance(arg, six.string_types):
            options = [
                initializers.KaimingNormal,
                initializers.NoOp,
            ]
            cls = {c.__name__: c for c in options}[arg]
        else:
            cls = arg
        return cls

    rectified = _rectify_class(arg, kw, _lookup)
    return rectified


def _rectify_monitor(arg, kw):
    def _lookup(arg):
        if isinstance(arg, six.string_types):
            options = []
            cls = {c.__name__: c for c in options}[arg]
        else:
            cls = arg
        return cls
    rectified = _rectify_class(arg, kw, _lookup)
    return rectified


def _rectify_dynamics(arg, kw):
    """
    Special params that control the dynamics of learning at the harness level
    at point that doesnt correspond to a decoupled class component.
    """
    if arg is None:
        arg = {}
    arg = arg.copy()
    dynamics = {
        # batch_step simulates larger batch sizes
        'batch_step': arg.pop('batch_step', 1),
        # Clips gradients
        'grad_norm_max': arg.pop('grad_norm_max', None),
        'grad_norm_type': arg.pop('grad_norm_type', 2),

        'warmup_iters': arg.pop('warmup_iters', None),  # HACKED AND EXPERIMENTAL
        'warmup_ratio': arg.pop('warmup_ratio', 1.0 / 3.0),  # HACKED AND EXPERIMENTAL
    }
    if not isinstance(dynamics['batch_step'], int):
        raise ValueError('batch_step must be an integer')
    if arg:
        raise KeyError('UNKNOWN dynamics: {}'.format(arg))
    return dynamics


def _rectify_model(arg, kw):
    if arg is None:
        return _rectify_class(None, kw)

    def _lookup_model(arg):
        import torchvision
        if isinstance(arg, six.string_types):
            options = [
                torchvision.models.AlexNet,
                torchvision.models.DenseNet,
            ]
            cls = {c.__name__: c for c in options}[arg]
        else:
            cls = arg
        return cls

    # Unwrap the model if was mounted
    if isinstance(arg, device.MountedModel):
        arg = arg.module

    rectified = _rectify_class(arg, kw, _lookup_model)
    return rectified


def _rectify_loaders(arg, kw):
    """
    Loaders are handled slightly differently than other classes
    We construct them eagerly (if they are not already constructed)
    """
    if arg is None:
        arg = {}

    loaders = None

    if isinstance(arg, dict):
        # Check if all args are already data loaders
        # if isinstance(arg.get('train', None), torch_data.DataLoader):
        if len(arg) and all(isinstance(v, torch_data.DataLoader) for v in arg.values()):
            # loaders were custom specified
            loaders = arg
            # TODO: extract relevant loader params efficiently
            cls = None
            if 'train' in loaders:
                kw2 = {
                    'batch_size': loaders['train'].batch_sampler.batch_size,
                }
            else:
                kw2 = {}
        else:
            # loaders is kwargs for `torch_data.DataLoader`
            arg = (torch_data.DataLoader, arg)
            # cls, kw2 = _rectify_class(None, arg, kw)
            rectified = _rectify_class(None, arg, kw)
            cls = rectified['cls']
            kw2 = rectified['cls_kw']
    else:
        raise ValueError('Loaders should be a dict')

    kwnice = ub.dict_subset(kw2, ['batch_size'], default=None)
    return loaders, cls, kw2, kwnice


class HyperParams(object):
    """
    Holds hyperparams relavent to training strategy

    The idea is that you tell it what is relevant FOR YOU, and then it makes
    you nice ids based on that. If you give if enough info it also allows you
    to use the training harness.

    CommandLine:
        python -m netharn.hyperparams HyperParams

    Example:
        >>> from netharn.hyperparams import *
        >>> hyper = HyperParams(
        >>>     criterion=('CrossEntropyLoss', {
        >>>         'weight': torch.FloatTensor([0, 2, 1]),
        >>>     }),
        >>>     optimizer=(torch.optim.SGD, {
        >>>         'nesterov': True, 'weight_decay': .0005,
        >>>         'momentum': 0.9, 'lr': .001,
        >>>     }),
        >>>     scheduler=('ReduceLROnPlateau', {}),
        >>> )
        >>> # xdoctest: +IGNORE_WANT
        >>> print(hyper.hyper_id())
        NoOp,SGD,dampening=0,lr=0.001,momentum=0.9,nesterov=True,weight_decay=0.0005,ReduceLROnPlateau,cooldown=0,eps=1e-08,factor=0.1,min_lr=0,mode=min,patience=10,threshold=0.0001,threshold_mode=rel,verbose=False,CrossEntropyLoss,ignore_index=-100,reduce=None,reduction=mean,size_average=None,weight=[0.0,2.0,1.0],DataLoader,batch_size=1,Dynamics,batch_step=1,grad_norm_max=None
    """

    def __init__(hyper,
                 # ----
                 datasets=None,
                 name=None,
                 workdir=None,
                 xpu=None,
                 loaders=None,
                 # ----
                 model=None,
                 criterion=None,
                 optimizer=None,
                 initializer=None,
                 scheduler=None,
                 # ---
                 dynamics=None,
                 monitor=None,
                 augment=None,
                 other=None,  # incorporated into the hash
                 extra=None,  # ignored when computing the hash
                 nice=None,  # alias of name
                 ):
        kwargs = {}

        hyper.datasets = datasets
        if name is None:
            import warnings
            warnings.warn('Specify "name" instead of "nice"')
            name = nice
        hyper.name = name
        hyper.workdir = workdir
        hyper.xpu = xpu

        loaders, cls, kw, kwnice = _rectify_loaders(loaders, kwargs)
        hyper.loaders = loaders
        hyper.loader_cls = cls
        hyper.loader_params = kw
        hyper.loader_params_nice = kwnice

        hyper._model_info = _rectify_model(model, kwargs)
        hyper.model_cls = hyper._model_info['cls']
        hyper.model_params = hyper._model_info['cls_kw']

        hyper._optimizer_info = _rectify_optimizer(optimizer, kwargs)
        hyper.optimizer_cls = hyper._optimizer_info['cls']
        hyper.optimizer_params = hyper._optimizer_info['cls_kw']

        hyper._scheduler_info = _rectify_lr_scheduler(scheduler, kwargs)
        hyper.scheduler_cls = hyper._scheduler_info['cls']
        hyper.scheduler_params = hyper._scheduler_info['cls_kw']

        hyper._criterion_info = _rectify_criterion(criterion, kwargs)
        hyper.criterion_cls = hyper._criterion_info['cls']
        hyper.criterion_params = hyper._criterion_info['cls_kw']

        hyper._initializer_info = _rectify_initializer(initializer, kwargs)
        hyper.initializer_cls = hyper._initializer_info['cls']
        hyper.initializer_params = hyper._initializer_info['cls_kw']

        hyper._monitor_info = _rectify_monitor(monitor, kwargs)
        hyper.monitor_cls = hyper._monitor_info['cls']
        hyper.monitor_params = hyper._monitor_info['cls_kw']

        hyper.dynamics = _rectify_dynamics(dynamics, kw)

        hyper.augment = augment
        hyper.other = other
        hyper.extra = extra

    @property
    def nice(hyper):
        """ alias of name for backwards compatibility """
        return hyper.name

    def make_model(hyper):
        """ Instanciate the model defined by the hyperparams """
        if hyper._model_info['instance'] is not None:
            return hyper._model_info['instance']
        model = hyper.model_cls(**hyper.model_params)
        return model

    def make_optimizer(hyper, parameters):
        """ Instanciate the optimizer defined by the hyperparams """
        if hyper._optimizer_info['instance'] is not None:
            return hyper._optimizer_info['instance']
        # What happens if we want to group parameters
        optimizer = hyper.optimizer_cls(parameters, **hyper.optimizer_params)
        return optimizer

    def make_scheduler(hyper, optimizer):
        """ Instanciate the lr scheduler defined by the hyperparams """
        if hyper._scheduler_info['instance'] is not None:
            return hyper._scheduler_info['instance']
        if hyper.scheduler_cls is None:
            return None
        kw = hyper.scheduler_params.copy()
        kw['optimizer'] = optimizer
        scheduler = hyper.scheduler_cls(**kw)
        return scheduler

    def make_initializer(hyper):
        """ Instanciate the initializer defined by the hyperparams """
        if hyper._initializer_info['instance'] is not None:
            return hyper._initializer_info['instance']
        initializer = hyper.initializer_cls(**hyper.initializer_params)
        return initializer

    def make_criterion(hyper):
        """ Instanciate the criterion defined by the hyperparams """
        if hyper._criterion_info['instance'] is not None:
            return hyper._criterion_info['instance']
        if hyper.criterion_cls is None:
            return None
        criterion = hyper.criterion_cls(**hyper.criterion_params)
        return criterion

    def make_loaders(hyper):
        if hyper.loaders is not None:
            return hyper.loaders
        else:
            loaders = {
                key: torch_data.DataLoader(dset, **hyper.loader_params)
                for key, dset in hyper.datasets.items()
            }
        return loaders

    def make_xpu(hyper):
        """ Instanciate the criterion defined by the hyperparams """
        xpu = device.XPU.coerce(hyper.xpu)
        return xpu

    def make_monitor(hyper):
        """ Instanciate the monitor defined by the hyperparams """
        if hyper._monitor_info['instance'] is not None:
            return hyper._monitor_info['instance']
        if hyper.monitor_cls is None:
            return None
        monitor = hyper.monitor_cls(**hyper.monitor_params)
        return monitor

    def other_id(hyper):
        """
            >>> from netharn.hyperparams import *
            >>> hyper = HyperParams(other={'augment': True, 'n_classes': 10, 'n_channels': 5})
            >>> hyper.hyper_id()
        """
        otherid = util.make_short_idstr(hyper.other, precision=4)
        return otherid

    def get_initkw(hyper):
        """
        Make list of class / params relevant to reproducing an experiment

        CommandLine:
            python ~/code/netharn/netharn/hyperparams.py HyperParams.get_initkw

        Example:
            >>> from netharn.hyperparams import *
            >>> hyper = HyperParams(
            >>>     criterion='CrossEntropyLoss',
            >>>     optimizer='Adam',
            >>>     loaders={'batch_size': 64},
            >>> )
            >>> print(ub.repr2(hyper.get_initkw()))
        """
        initkw = OrderedDict()
        def _append_part(key, cls, params, initkw):
            """
            append an id-string derived from the class and params.
            TODO: what if we have an instance and not a cls/params tuple?
            """
            if cls is None:
                initkw[key] = None
            else:
                d = OrderedDict()
                for k, v in sorted(params.items()):
                    # if k in total:
                    #     raise KeyError(k)
                    if isinstance(v, torch.Tensor):
                        v = v.numpy()
                    if isinstance(v, np.ndarray):
                        if v.dtype.kind == 'f':
                            try:
                                v = list(map(float, v))
                            except Exception:
                                v = v.tolist()
                        else:
                            raise NotImplementedError()
                    d[k] = v
                    # total[k] = v
                if isinstance(cls, six.string_types):
                    type_str = cls
                else:
                    modname = cls.__module__
                    type_str = modname + '.' + cls.__name__
                # param_str = util.make_idstr(d)
                initkw[key] = (type_str, d)

        _append_part('model', hyper.model_cls, hyper.model_params, initkw)
        _append_part('initializer', hyper.initializer_cls, hyper.initializer_params, initkw)
        _append_part('optimizer', hyper.optimizer_cls, hyper.optimizer_params, initkw)
        _append_part('scheduler', hyper.scheduler_cls, hyper.scheduler_params, initkw)
        _append_part('criterion', hyper.criterion_cls, hyper.criterion_params, initkw)

        # TODO: should other be included in initkw? I think it should.
        # probably should also include monitor, xpu, nice

        # Loader is a bit hacked
        _append_part('loader', hyper.loader_cls, hyper.loader_params_nice, initkw)
        _append_part('dynamics', 'Dynamics', hyper.dynamics, initkw)

        return initkw

    def augment_json(hyper):
        """
        Get augmentation info in json format

        Example:
            >>> from netharn.hyperparams import *
            >>> import imgaug
            >>> augment = imgaug.augmenters.Affine()
            >>> hyper = HyperParams(augment=augment)
            >>> info = hyper.augment_json()
            >>> assert info['__class__'] == 'Affine'
            >>> hyper = HyperParams(augment=OrderedDict())
            >>> assert hyper.augment_json() == {}
        """
        if hyper.augment is None:
            return None
        elif imgaug is not None and isinstance(hyper.augment, imgaug.augmenters.Augmenter):
            from netharn.data.transforms.augmenter_base import ParamatarizedAugmenter
            augment_json = ParamatarizedAugmenter._json_id(hyper.augment)
        elif isinstance(hyper.augment, six.string_types):
            return hyper.augment
        # Some classes in imgaug (e.g. Sequence) inherit from list,
        # so we have to check for Augmenter before we check for list type
        # if isinstance(hyper.augment, (dict, list)):
        elif isinstance(hyper.augment, OrderedDict):
            # dicts are specified in json format
            try:
                # hashable data should be loosely json-compatible
                _hash_data(hyper.augment)
            except TypeError:
                raise TypeError(
                    'NOT IN ORDERED JSON FORMAT hyper.augment={}'.format(
                        hyper.augment))
            augment_json = hyper.augment
        else:
            raise TypeError('Specify augment in json format')
        return augment_json

    def input_id(hyper, short=False, hashed=False):
        pass

    def _parts_id(hyper, parts, short=False, hashed=False):
        id_parts = []
        for key, value in parts.items():
            if value is None:
                continue
            clsname, params = value
            type_str = clsname.split('.')[-1]
            id_parts.append(type_str)

            # Precidence of specifications (from lowest to highest)
            # SF=single flag, EF=explicit flag
            # SF-short, SF-hash, EF-short EF-hash
            request_short = short is True
            request_hash = hashed is True
            if (ub.iterable(short) and key in short):
                request_hash = False
                request_short = True
            if (ub.iterable(hashed) and key in hashed):
                request_hash = True
                request_short = False

            if request_hash:
                param_str = util.make_idstr(params)
                param_str = _hash_data(param_str)[0:6]
            elif request_short:
                param_str = util.make_short_idstr(params)
            else:
                param_str = util.make_idstr(params)

            if param_str:
                id_parts.append(param_str)
        idstr = ','.join(id_parts)
        return idstr

    def hyper_id(hyper, short=False, hashed=False):
        """
        Identification string that uniquely determined by training hyper.
        Suitable for hashing.

        Note:
            setting short=True is deprecated

        CommandLine:
            python -m netharn.hyperparams HyperParams.hyper_id

        Example:
            >>> from netharn.hyperparams import *
            >>> hyper = HyperParams(criterion='CrossEntropyLoss', other={'n_classes': 10, 'n_channels': 5})
            >>> print(hyper.hyper_id())
            >>> print(hyper.hyper_id(hashed=True))
            >>> #print(hyper.hyper_id(short=['optimizer']))
            >>> #print(hyper.hyper_id(short=['optimizer'], hashed=True))
            >>> #print(hyper.hyper_id(short=['optimizer', 'criterion'], hashed=['criterion']))
        """
        parts = hyper.get_initkw()
        return hyper._parts_id(parts, short, hashed)

    def train_info(hyper, train_dpath=None):
        """
        Create json metadata that details enough information such that it would
        be possible for a human to reproduce the experiment.

        Example:
            >>> import netharn as nh
            >>> datasets = {
            >>>     'train': nh.data.ToyData2d(size=3, border=1, n=256, rng=0),
            >>>     'vali': nh.data.ToyData2d(size=3, border=1, n=128, rng=1),
            >>> }
            >>> hyper = nh.hyperparams.HyperParams(**{
            >>>     # --- Data First
            >>>     'datasets'    : datasets,
            >>>     'nice'        : 'demo',
            >>>     'workdir'     : ub.ensure_app_cache_dir('netharn/demo'),
            >>>     'loaders'     : {'batch_size': 64},
            >>>     'xpu'         : nh.XPU.coerce('auto'),
            >>>     # --- Algorithm Second
            >>>     'model'       : (nh.models.ToyNet2d, {}),
            >>>     'optimizer'   : (nh.optimizers.SGD, {
            >>>         'lr': 0.001
            >>>     }),
            >>>     'criterion'   : (nh.criterions.CrossEntropyLoss, {}),
            >>>     #'criterion'   : (nh.criterions.FocalLoss, {}),
            >>>     'initializer' : (nh.initializers.KaimingNormal, {
            >>>         'param': 0,
            >>>     }),
            >>>     'scheduler'   : (nh.schedulers.ListedLR, {
            >>>         'step_points': {0: .001, 2: .01, 5: .015, 6: .005, 9: .001},
            >>>         'interpolate': True,
            >>>     }),
            >>>     'monitor'     : (nh.Monitor, {
            >>>         'max_epoch': 10
            >>>     }),
            >>> })
            >>> info = hyper.train_info()
            >>> print(ub.repr2(info))
        """
        given_explicit_train_dpath = train_dpath is not None
        # TODO: needs MASSIVE cleanup and organization

        # TODO: if pretrained is another netharn model, then we should read that
        # train_info if it exists and append it to a running list of train_info

        if hyper.model_cls is None:
            # import utool
            # utool.embed()
            raise ValueError('model_cls is None')
        # arch = hyper.model_cls.__name__

        train_dset = hyper.datasets.get('train', None)
        if train_dset is not None and hasattr(train_dset, 'input_id'):
            input_id = train_dset.input_id
            if callable(input_id):
                input_id = input_id()
        else:
            warnings.warn(
                'FitHarn cannot track the training dataset state because '
                'harn.datasets["train"] is missing the "input_id" attribute.'
            )
            input_id = 'none'

        def _hash_data(data):
            return ub.hash_data(data, hasher='sha512', base='abc', types=True)

        train_hyper_id_long = hyper.hyper_id()
        train_hyper_id_brief = hyper.hyper_id(short=False, hashed=True)
        train_hyper_hashid = _hash_data(train_hyper_id_long)[:8]

        # TODO: hash this to some degree
        other_id = hyper.other_id()

        augment_json = hyper.augment_json()

        aug_brief = 'AU' + _hash_data(augment_json)[0:6]
        # extra_hash = _hash_data([hyper.centering])[0:6]

        train_id = '{}_{}_{}_{}'.format(
            _hash_data(input_id)[:6], train_hyper_id_brief,
            aug_brief, other_id)

        # Gather all information about this run into a single hash

        """
        NOTE:
            On choosing the length to truncate the hash.

            If we have an alphabet of size A=26, and we truncate to M=8
            samples, then the number of possible hash values is N = A ** M.
            The probability we will have a collision (assuming an ideal hash
            function where all outputs are equally likely) in r different
            inputs is given by the following function. Note this is the
            birthday paradox problem [1].


            ```python
            from scipy import exp, log
            from scipy.special import gammaln
            def prob_unique(N, r):
                return exp( gammaln(N+1) - gammaln(N-r+1) - r*log(N) )

            A = 26  # size of the alphabet for _hash_data
            M = 8   # number of characters we truncate at
            N = A ** M  # number of possible hash values

            r = 1000

            prob_collision = 1 - prob_unique(N, r)
            print('prob_collision = {!r}'.format(prob_collision))
            ```

            This is approximately 0.00056 or about 1 in 1784.
            When r = 10000, it becomes had to compute the number because of
            floating point errors, but the probability is likely astronomically
            low. I doubt we will ever run training in the same work directory
            (and with the same nice name) 10,000 different times, so using an 8
            character hash seems safe and user friendly for this purpose.
            Perhaps we may move to 12, 16, or 32+ in the future, but for the
            pre 1.0 netharn, 8 seems fine.

            References:
                ..[1] https://www.johndcook.com/blog/2016/01/30/general-birthday-problem/

        """
        train_hashid = _hash_data(train_id)[0:8]

        name = hyper.name

        nice_dpath = None
        if not given_explicit_train_dpath:
            # setup a cannonical and a linked symlink dir
            train_dpath = normpath(
                    join(hyper.workdir, 'fit', 'runs', name, train_hashid))
            # also setup a "nice" custom name, which may conflict, but oh well
            if name:
                try:
                    nice_dpath = normpath(
                            join(hyper.workdir, 'fit', 'nice', name))
                except Exception:
                    print('hyper.workdir = {!r}'.format(hyper.workdir))
                    print('hyper.name = {!r}'.format(hyper.name))
                    raise

        # make temporary initializer so we can infer the history
        temp_initializer = hyper.make_initializer()
        init_history = temp_initializer.history()

        train_info =  ub.odict([
            ('train_hashid', train_hashid),

            ('train_id', train_id),

            ('workdir', hyper.workdir),

            ('aug_brief', aug_brief),

            ('input_id', input_id),

            ('other_id', other_id),

            ('hyper', hyper.get_initkw()),

            ('train_hyper_id_long', train_hyper_id_long),
            ('train_hyper_id_brief', train_hyper_id_brief),
            ('train_hyper_hashid', train_hyper_hashid),
            ('init_history', init_history),
            ('init_history_hashid', _hash_data(util.make_idstr(init_history))),

            ('nice', hyper.name),

            ('old_train_dpath', normpath(
                join(hyper.workdir, 'fit', 'runs', train_hashid))),

            ('train_dpath', train_dpath),
            # ('link_dpath', link_dpath),
            ('nice_dpath', nice_dpath),

            ('given_explicit_train_dpath', given_explicit_train_dpath),

            # TODO, add in n_classes if applicable
            # TODO, add in centering if applicable
            # ('centering', hyper.centering),

            ('other', hyper.other),

            # HACKED IN
            ('augment', hyper.augment_json()),

            ('extra', hyper.extra),

            ('argv', sys.argv),
            ('hostname', platform.node()),
        ])
        return train_info

    @classmethod
    def demo(HyperParams):
        import netharn as nh
        hyper = HyperParams(**{
            # ================
            # Environment Components
            'workdir'     : ub.ensure_app_cache_dir('netharn/tests/demo'),
            'name'        : 'demo',
            'xpu'         : nh.XPU.coerce('argv'),
            # workdir is a directory where intermediate results can be saved
            # nice symlinks <workdir>/fit/nice/<nice> -> ../runs/<hashid>
            # XPU auto select a gpu if idle and VRAM>6GB else a cpu
            # ================
            # Data Components
            'datasets'    : {  # dict of plain ol torch.data.Dataset instances
                'train': nh.data.ToyData2d(size=3, border=1, n=256, rng=0),
                'vali': nh.data.ToyData2d(size=3, border=1, n=128, rng=1),
                'test': nh.data.ToyData2d(size=3, border=1, n=128, rng=1),
            },
            'loaders'     : {'batch_size': 64},  # DataLoader instances or kw
            # ================
            # Algorithm Components
            # Note the (cls, kw) tuple formatting
            'model'       : (nh.models.ToyNet2d, {}),
            'optimizer'   : (nh.optimizers.SGD, {
                'lr': 0.0001
            }),
            # focal loss is usually better than nh.criterions.CrossEntropyLoss
            'criterion'   : (nh.criterions.FocalLoss, {}),
            'initializer' : (nh.initializers.KaimingNormal, {
                'param': 0,
            }),
            # these may receive an overhaul soon
            'scheduler'   : (nh.schedulers.ListedLR, {
                'points': {0: .0001, 2: .01, 5: .015, 6: .005, 9: .001},
                'interpolate': True,
            }),
            'monitor'     : (nh.Monitor, {
                'max_epoch': 10,
            }),
            # dynamics are a config option that modify the behavior of the main
            # training loop. These parameters effect the learned model.
            'dynamics'   : {'batch_step': 4},
        })
        return hyper

if __name__ == '__main__':
    r"""
    CommandLine:
        python -m netharn.hyperparams
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



import six


def default_kwargs(cls):
    """
    Grab initkw defaults from the constructor

    Args:
        cls (type | callable): a class or function

    Example:
        >>> from netharn.util.util_inspect import *  # NOQA
        >>> import netharn as nh
        >>> import torch
        >>> import ubelt as ub
        >>> cls = torch.optim.Adam
        >>> default_kwargs(cls)
        >>> cls = nh.initializers.KaimingNormal
        >>> print(ub.repr2(default_kwargs(cls), nl=0))
        {'mode': 'fan_in', 'param': 0}
        >>> cls = nh.initializers.NoOp
        >>> default_kwargs(cls)
        {}

    SeeAlso:
        xinspect.get_func_kwargs(cls)
    """
    if six.PY2:
        if cls.__init__ is object.__init__:
            # hack for python2 classes without __init__
            return {}
        else:
            import funcsigs
            sig = funcsigs.signature(cls)
    else:
        import inspect
        sig = inspect.signature(cls)

    default_kwargs = {
        k: p.default
        for k, p in sig.parameters.items()
        if p.default is not p.empty
    }
    return default_kwargs



#
#   Darknet YOLOv2 model
#   Copyright EAVISE
#   Modified for netharn
#
from collections import OrderedDict
import numpy as np
import torch
import torch.nn as nn
import ubelt as ub
from netharn import layers
from distutils.version import LooseVersion


_TORCH_HAS_BOOL_COMP = LooseVersion(torch.__version__) >= LooseVersion('1.2.0')


class Conv2dBatchLeaky(layers.Module):
    """ This convenience layer groups a 2D convolution, a batchnorm and a leaky ReLU.
    They are executed in a sequential manner.

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int or tuple): Size of the kernel of the convolution
        stride (int or tuple): Stride of the convolution
        padding (int or tuple): padding of the convolution
        leaky_slope (number, optional): Controls the angle of the negative slope of the leaky ReLU; Default **0.1**

    Example:
        >>> self = Conv2dBatchLeaky(1, 1, 3, 1, 0)
        >>> self.output_shape_for((1, 1, 3, 3))
        >>> self.receptive_field_for()
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride, padding, leaky_slope=0.1):
        super(Conv2dBatchLeaky, self).__init__()

        # Parameters
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.leaky_slope = leaky_slope

        # Layer
        self.layers = layers.Sequential(
            nn.Conv2d(self.in_channels, self.out_channels, self.kernel_size, self.stride, self.padding, bias=False),
            nn.BatchNorm2d(self.out_channels),
            nn.LeakyReLU(self.leaky_slope, inplace=True)
        )

    def __repr__(self):
        s = '{name} ({in_channels}, {out_channels}, kernel_size={kernel_size}, stride={stride}, padding={padding}, negative_slope={leaky_slope})'
        return s.format(name=self.__class__.__name__, **self.__dict__)

    def forward(self, x):
        x = self.layers(x)
        return x

    def output_shape_for(self, input_shape):
        return self.layers.output_shape_for(input_shape)

    def receptive_field_for(self, input_field=None):
        return self.layers.receptive_field_for(input_field)


class Reorg(nn.Module):
    """
    This layer reorganizes a tensor according to a stride.

    The input must have 4 dimensions: [B, C, H, W]. The H and W dimesions will
    be sliced by the stride and then stacked in dimension C. There is some
    weirdness about how this works see [1] for details.

    References:
        [1] https://leimao.github.io/blog/Reorg-Layer-Explained/

    Args:
        stride (int): stride to divide the input tensor
        darknet (int): use original (slightly odd) darknet compatible behavior

    Example:
        >>> from netharn.models.yolo2.yolo2 import *  # NOQA
        >>> self = Reorg(2)
        >>> x = torch.arange(64).view(1, 4, 4, 4)
        >>> y1 = self.forward(x)
        >>> self.output_shape_for((1, 8, 32, 32))
        >>> self.output_shape_for((1, 8, 16, 32))

    Ignore:
        make = lambda *a: torch.arange(int(np.prod(a))).view(*a)
        x = make(1, 4, 4, 4)
        self.darknet = True
        y1 = self.forward(x)
        self.darknet = False
        y2 = self.forward(x)
        print(ub.hzcat(['y1 = {!r}'.format(y1), ' | ', 'y2 = {!r}'.format(y2)]))
    """
    def __init__(self, stride=2, darknet=True):
        super(Reorg, self).__init__()
        if not isinstance(stride, int):
            raise TypeError('stride is not an int [{}]'.format(type(stride)))
        self.stride = stride
        self.darknet = darknet

    def __repr__(self):
        return '{} (stride={}, darknet_compatible_mode={})'.format(
            self.__class__.__name__, self.stride, self.darknet)

    def forward(self, x):
        assert(x.data.dim() == 4)
        B, C, H, W = x.shape

        if H % self.stride != 0:
            raise ValueError('Dimension mismatch: {} is not divisible by {}'.format(W, self.stride))
        if W % self.stride != 0:
            raise ValueError('Dimension mismatch: {} is not divisible by {}'.format(W, self.stride))

        # darknet compatible version from: https://github.com/thtrieu/darkflow/issues/173#issuecomment-296048648
        if self.darknet:
            x = x.view(B, C // (self.stride ** 2), H, self.stride, W, self.stride).contiguous()
            x = x.permute(0, 3, 5, 1, 2, 4).contiguous()
            x = x.view(B, -1, H // self.stride, W // self.stride)
        else:
            ws, hs = self.stride, self.stride
            x = x.view(B, C, H // hs, hs, W // ws, ws).transpose(3, 4).contiguous()
            x = x.view(B, C, H // hs * W // ws, hs * ws).transpose(2, 3).contiguous()
            x = x.view(B, C, hs * ws, H // hs, W // ws).transpose(1, 2).contiguous()
            x = x.view(B, hs * ws * C, H // hs, W // ws)
        return x

    def output_shape_for(self, input_shape):
        B, C, H, W = input_shape

        C_new = C * int(self.stride ** 2)
        H_new = H // (self.stride)
        W_new = W // (self.stride)

        output_shape = (B, C_new, H_new, W_new)
        return output_shape


class Yolo2(layers.AnalyticModule):
    """
    `Yolov2`_ implementation with pytorch.

    Modified version original taken from lightnet

    Integrated netharn module mixins

    Example:
        >>> # xdoc: +REQUIRES(--download, module:ndsampler)
        >>> from netharn.models.yolo2.yolo2 import *
        >>> torch.random.manual_seed(0)
        >>> B, C, Win, Hin = 2, 20, 96, 96
        >>> self = Yolo2(C, conf_thresh=4.9e-2)
        >>> im_data = torch.randn(B, 3, Hin, Win)
        >>> # the _forward function produces raw YOLO output
        >>> network_output = self.forward(im_data)
        >>> A = len(self.anchors)
        >>> Wout, Hout = Win // 32, Hin // 32
        >>> # The default coder function will construct the bounding boxes
        >>> # Each item in `batch_dets` is a list of Detections objects.
        >>> batch_dets = self.coder.decode_batch(network_output)
        >>> boxes = batch_dets[0].numpy().take([0, 1])
        >>> print(ub.repr2(boxes.data, nl=1))  # xdoc: +IGNORE_WANT
    """

    def __init__(self, classes, conf_thresh=.25, nms_thresh=.4,
                 in_channels=3, anchors=None, input_stats=None):
        """ Network initialisation """
        super(Yolo2, self).__init__()

        if anchors is None:
            anchors = np.array([(1.3221, 1.73145), (3.19275, 4.00944),
                                (5.05587, 8.09892), (9.47112, 4.84053),
                                (11.2364, 10.0071)], dtype=np.float)

        import ndsampler
        classes = ndsampler.CategoryTree.coerce(classes)

        if input_stats is None:
            input_stats = {}
        self.input_norm = layers.InputNorm(**input_stats)

        self.classes = classes
        self.num_classes = len(classes)
        self.anchors = anchors

        # Network
        layer_list = [
            # Sequence 0 : input = image tensor
            OrderedDict([
                ('1_convbatch',     Conv2dBatchLeaky(in_channels, 32, 3, 1, 1)),
                ('2_max',           nn.MaxPool2d(2, 2)),
                ('3_convbatch',     Conv2dBatchLeaky(32, 64, 3, 1, 1)),
                ('4_max',           nn.MaxPool2d(2, 2)),
                ('5_convbatch',     Conv2dBatchLeaky(64, 128, 3, 1, 1)),
                ('6_convbatch',     Conv2dBatchLeaky(128, 64, 1, 1, 0)),
                ('7_convbatch',     Conv2dBatchLeaky(64, 128, 3, 1, 1)),
                ('8_max',           nn.MaxPool2d(2, 2)),
                ('9_convbatch',     Conv2dBatchLeaky(128, 256, 3, 1, 1)),
                ('10_convbatch',    Conv2dBatchLeaky(256, 128, 1, 1, 0)),
                ('11_convbatch',    Conv2dBatchLeaky(128, 256, 3, 1, 1)),
                ('12_max',          nn.MaxPool2d(2, 2)),
                ('13_convbatch',    Conv2dBatchLeaky(256, 512, 3, 1, 1)),
                ('14_convbatch',    Conv2dBatchLeaky(512, 256, 1, 1, 0)),
                ('15_convbatch',    Conv2dBatchLeaky(256, 512, 3, 1, 1)),
                ('16_convbatch',    Conv2dBatchLeaky(512, 256, 1, 1, 0)),
                ('17_convbatch',    Conv2dBatchLeaky(256, 512, 3, 1, 1)),
            ]),

            # Sequence 1 : input = sequence0
            OrderedDict([
                ('18_max',          nn.MaxPool2d(2, 2)),
                ('19_convbatch',    Conv2dBatchLeaky(512, 1024, 3, 1, 1)),
                ('20_convbatch',    Conv2dBatchLeaky(1024, 512, 1, 1, 0)),
                ('21_convbatch',    Conv2dBatchLeaky(512, 1024, 3, 1, 1)),
                ('22_convbatch',    Conv2dBatchLeaky(1024, 512, 1, 1, 0)),
                ('23_convbatch',    Conv2dBatchLeaky(512, 1024, 3, 1, 1)),
                ('24_convbatch',    Conv2dBatchLeaky(1024, 1024, 3, 1, 1)),
                ('25_convbatch',    Conv2dBatchLeaky(1024, 1024, 3, 1, 1)),
            ]),

            # Sequence 2 : input = sequence0
            OrderedDict([
                ('26_convbatch',    Conv2dBatchLeaky(512, 64, 1, 1, 0)),
                ('27_reorg',        Reorg(2)),
            ]),

            # Sequence 3 : input = sequence2 + sequence1
            OrderedDict([
                ('28_convbatch',    Conv2dBatchLeaky((4 * 64) + 1024, 1024, 3, 1, 1)),
                ('29_conv',         nn.Conv2d(1024, len(self.anchors) * (5 + self.num_classes), 1, 1, 0)),
            ])
        ]
        self.layers = nn.ModuleList(
            [nn.Sequential(layer_dict) for layer_dict in layer_list])

        self.reduction = 32   # = 2 ** 5 =  2 ** (num_downsample_ops)
        self.coder = YoloCoder(
            self.classes, self.anchors, conf_thresh, nms_thresh,
            reduction=self.reduction,
        )

    def forward(self, inputs):
        """
        CommandLine:
            xdoctest -m netharn.models.yolo2.yolo2 Yolo2.forward

        Example:
            >>> # DISABLE_DOCTEST
            >>> # xdoc: +REQUIRES(--download, module:ndsampler)
            >>> from netharn.models.yolo2.yolo2 import *
            >>> self = Yolo2.demo()
            >>> inputs = self.demo_inputs()
            >>> output = self(inputs)
            >>> batch_dets = self.coder.decode_batch(output)
            >>> dets = batch_dets[0]
            >>> # xdoc: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()  # xdoc: +SKIP
            >>> kwplot.imshow(inputs[0], colorspace='rgb', fnum=1, doclf=True)
            >>> dets.scale(inputs.shape[-2:][::-1]).draw()
            >>> kwplot.show_if_requested()
        """
        normed = self.input_norm(inputs)
        out0 = self.layers[0](normed)
        out1 = self.layers[1](out0)
        # Route : layers=-9
        out2 = self.layers[2](out0)
        # Route : layers=-1,-4
        combo = torch.cat((out2, out1), 1)
        raw0 = self.layers[3](combo)

        # Reshape output to separate the anchor dimension from the box and class dimension
        nB, nO, nH, nW = raw0.shape
        nA = len(self.anchors)
        nC = self.num_classes
        raw = raw0.view(nB, nA, 5 + nC, nH, nW)
        output = {
            'cxywh_energy': raw[:, :, 0:4, :, :],
            'score_energy': raw[:, :, 4:5, :, :],
            'class_energy': raw[:, :, 5:,  :, :],
        }
        return output

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        Define forward in a symbolic way such that output shape and receptive
        field can be analytically computed.

        Ignore:
            >>> import netharn as nh
            >>> globals().update(nh.layers.AnalyticModule._analytic_shape_kw())
            >>> inputs = [2, 3, 288, 288]
            >>> self = Yolo2.demo()
            >>> out = self._analytic_forward(inputs, _OutputFor, _Output, _Hidden)
            >>> print(ub.repr2(out.hidden.shallow(2), nl=-1))
        """
        hidden = _Hidden()
        normed = hidden['input_norm'] = _OutputFor(self.input_norm)(inputs)
        # Preproc
        out0 = hidden['out0'] = _OutputFor(self.layers[0])(normed)

        # High level and reorg branches
        out1 = hidden['out1'] = _OutputFor(self.layers[1])(out0)
        out2 = hidden['out2'] = _OutputFor(self.layers[2])(out0)

        combo = hidden['combo'] = _OutputFor(torch.cat)((out2, out1), 1)

        raw0 = hidden['raw0'] = _OutputFor(self.layers[3])(combo)

        nA = len(self.anchors)
        nC = self.num_classes
        nB, nO, nH, nW = _OutputFor.shape(raw0)

        raw = hidden['raw'] = _OutputFor.view(raw0, nB, nA, 5 + nC, nH, nW)

        output = hidden['output'] = {
            'cxywh_energy': _OutputFor.getitem(raw)[:, :, 0:4, :, :],
            'score_energy': _OutputFor.getitem(raw)[:, :, 4:5, :, :],
            'class_energy': _OutputFor.getitem(raw)[:, :, 5:,  :, :],
        }
        return _Output.coerce(output, hidden)

    # def output_shape_for(self, input_shape):
    #     """
    #     Shape of the output produced by this network
    #     """
    #     nB, c, inH, inW = input_shape
    #     outH = inH / self.factor
    #     outW = inW / self.factor
    #     nA = len(self.anchors)
    #     nC = self.num_classes
    #     return (nB, nA, 5 + nC, outH, outW)

    @classmethod
    def demo(cls, key='lightnet'):

        classes = ['aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus',
                   'car', 'cat', 'chair', 'cow', 'diningtable', 'dog', 'horse',
                   'motorbike', 'person', 'pottedplant', 'sheep', 'sofa',
                   'train', 'tvmonitor']
        self = Yolo2(classes=classes, conf_thresh=0.01, nms_thresh=0.4)

        import netharn as nh
        weights_fpath = demo_voc_weights(key)
        initializer = nh.initializers.Pretrained(weights_fpath)
        init_info = initializer(self, verbose=0)  # NOQA

        self.eval()
        return self

    def demo_inputs(self):
        inp_size = (288, 288)
        im_data, rgb255 = demo_image(inp_size)
        inputs = torch.cat([im_data, im_data])  # make a batch size of 2
        return inputs


class YoloCoder(object):
    """
    Translate yolo inputs and outputs between input and output space.

    Example:
        >>> # xdoctest: +REQUIRES(--download)
        >>> from netharn.models.yolo2.yolo2 import *  # NOQA
        >>> self = YoloCoder.demo()
        >>> print(self.__json__())
    """
    _coder_version = 3

    def __init__(self, classes, anchors, conf_thresh=0.001, nms_thresh=0.4,
                 reduction=32):
        import ndsampler
        self.classes = ndsampler.CategoryTree.coerce(classes)
        self.num_anchors = len(anchors)
        self.anchor_step = len(anchors[0])
        self.anchors = torch.Tensor(anchors)
        self.conf_thresh = conf_thresh
        self.nms_thresh = nms_thresh

        self.reduction = reduction  # set based on model downsampling

    @classmethod
    def coerce(YoloCoder, arg):
        if isinstance(arg, dict):
            return YoloCoder(**arg)
        elif isinstance(arg, YoloCoder):
            return arg
        else:
            raise TypeError(type(arg))

    def __json__(self):
        return self.__getstate__()

    def __getstate__(self):
        data = self.__dict__.copy()
        data['classes'] = self.classes.__json__()
        data['anchors'] = self.anchors.data.cpu().numpy().tolist()
        data.pop('num_anchors')
        data.pop('anchor_step')
        return data

    def __setstate__(self, state):
        import ndsampler
        state['classes'] = ndsampler.CategoryTree.coerce(state['classes'])
        state['anchors'] = torch.Tensor(state['anchors'])
        state['num_anchors'] = len(state['anchors'])
        state['anchor_step'] = len(state['anchors'][0])
        self.__dict__.update(state)

    @classmethod
    def demo(cls):
        anchors = np.array([
            (1.3221, 1.73145), (3.19275, 4.00944), (5.05587, 8.09892),
            (9.47112, 4.84053), (11.2364, 10.0071)])
        self = cls(classes=20, anchors=anchors, conf_thresh=.14, nms_thresh=0.5)
        return self

    def demo_output(self, bsize=3):
        nA = len(self.anchors)
        nC = len(self.classes)
        raw = torch.randn(bsize, nA, 5 + nC, 9, 9)
        output = {
            'cxywh_energy': raw[:, :, 0:4, :, :],
            'score_energy': raw[:, :, 4:5, :, :],
            'class_energy': raw[:, :, 5:,  :, :],
        }
        return output

    def decode_batch(self, output, forloss=False):
        """
        Returns array of detections for every image in batch

        Example:
            >>> # xdoc: +REQUIRES(--download, module:ndsampler)
            >>> from netharn.models.yolo2.yolo2 import *  # NOQA
            >>> self = YoloCoder.demo()
            >>> output = self.demo_output()
            >>> batch_dets = self.decode_batch(output)
            >>> batch_dets = self.decode_batch(output, forloss=True)

        Example:
            >>> # xdoc: +REQUIRES(--download, module:ndsampler)
            >>> info = dev_demodata()
            >>> self, output = ub.take(info, ['coder', 'outputs'])
            >>> batch_dets = self.decode_batch(output)
            >>> dets = batch_dets[0]
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.figure(fnum=1, doclf=True)
            >>> kwplot.imshow(info['rgb255'], colorspace='rgb')
            >>> dets.scale(info['orig_sizes'][0]).draw()
            >>> kwplot.show_if_requested()
        """
        import kwimage
        # dont modify inplace
        # output = output.clone()

        class_energy = output['class_energy']
        score_energy = output['score_energy']
        cxywh_energy = output['cxywh_energy']

        # Variables
        nB = class_energy.shape[0]
        nH, nW = class_energy.shape[-2:]
        nA = self.num_anchors

        device = class_energy.device

        if self.anchors.device != device:
            self.anchors = self.anchors.to(device)

        # Compute xc,yc, nW,nH, box_score on Tensor
        lin_x = torch.linspace(0, nW - 1, nW, device=device).repeat(nH, 1)
        lin_y = torch.linspace(0, nH - 1, nH, device=device).repeat(nW, 1).t().contiguous()
        anchor_w = self.anchors[:, 0].contiguous().view(1, nA, 1, 1)
        anchor_h = self.anchors[:, 1].contiguous().view(1, nA, 1, 1)

        if forloss:
            # TODO : rectify
            coord = torch.empty_like(cxywh_energy)
            coord[:, :, 0:2, :, :] = cxywh_energy[:, :, 0:2, :, :].sigmoid()    # cx,cy
            coord[:, :, 2:4, :, :] = cxywh_energy[:, :, 2:4, :, :]              # w,h

            with torch.no_grad():
                pred_boxes = torch.empty_like(cxywh_energy, device=device).view(-1, 4)
                pred_boxes[:, 0] = (coord[:, :, 0, :, :] + lin_x).view(-1)
                pred_boxes[:, 1] = (coord[:, :, 1, :, :] + lin_y).view(-1)
                pred_boxes[:, 2] = (coord[:, :, 2, :, :].exp() * anchor_w).view(-1)
                pred_boxes[:, 3] = (coord[:, :, 3, :, :].exp() * anchor_h).view(-1)

            info = {
                'coord': coord,
                'pred_boxes': pred_boxes,
            }
            return info
        else:
            cxywh = cxywh_energy.clone()
            # cxywh_ = cxywh.view(nB, self.num_anchors, -1, nH,  nW)
            cxywh[:, :, 0, :].sigmoid_().add_(lin_x).div_(nW)          # X center
            cxywh[:, :, 1, :].sigmoid_().add_(lin_y).div_(nH)          # Y center
            cxywh[:, :, 2, :].exp_().mul_(anchor_w).div_(nW)           # Width
            cxywh[:, :, 3, :].exp_().mul_(anchor_h).div_(nH)           # Height

        score = score_energy.sigmoid()                             # Box score

        # Compute class_score
        if len(self.classes) > 1:
            cls_scores = torch.nn.functional.softmax(class_energy, dim=2)

            cls_max, cls_max_idx = torch.max(cls_scores, 2, keepdim=True)
            cls_max.mul_(score)
        else:
            cls_max = score
            cls_max_idx = torch.zeros_like(cls_max)

        # Save detection if conf*class_conf is higher than threshold
        flags = cls_max >= self.conf_thresh
        flags_flat = flags.view(-1)

        if flags.sum() == 0:
            batch_dets = []
            for i in range(nB):
                batch_dets.append(kwimage.Detections(
                    boxes=kwimage.Boxes(torch.empty((0, 4), dtype=torch.float32, device=device), 'cxywh'),
                    scores=torch.empty(0, dtype=torch.float32, device=device),
                    class_idxs=torch.empty(0, dtype=torch.int64, device=device),
                    classes=self.classes
                ))
        else:
            # Permute so the bbox dim (i.e. xywh) is trailing
            coords = cxywh.permute(0, 1, 3, 4, 2).contiguous().view(-1, 4)
            coords = coords[flags.view(-1)]

            scores = cls_max[flags]
            class_idxs = cls_max_idx[flags]

            stacked_dets = kwimage.Detections(
                boxes=kwimage.Boxes(coords, 'cxywh'),
                scores=scores,
                class_idxs=class_idxs,
                classes=self.classes
            )

            # Get indexes of splits between images of batch
            max_det_per_batch = len(self.anchors) * nH * nW
            m = max_det_per_batch
            flags_flat.int = flags_flat.int()
            slices = [slice(m * i, m * (i + 1)) for i in range(nB)]
            det_per_batch = torch.IntTensor([flags_flat[s].sum()
                                             for s in slices])
            split_idx = torch.cumsum(det_per_batch, dim=0)

            batch_dets = []
            start = 0
            for end in split_idx:
                dets = stacked_dets[start: end]
                dets = dets.non_max_supress(thresh=self.nms_thresh)
                batch_dets.append(dets)
                start = end
        return batch_dets


class YoloLoss(layers.common.Loss):
    """ Computes region loss from darknet network output and target annotation.

    Args:
        coder (YoloCoder): coder for the network

        # classes (int): number of categories
        # anchors (list): 2D list representing anchor boxes (see :class:`lightnet.network.Darknet`)

        coord_scale (float): weight of bounding box coordinates
        noobject_scale (float): weight of regions without target boxes
        object_scale (float): weight of regions with target boxes
        class_scale (float): weight of categorical predictions
        thresh (float): minimum iou between a predicted box and ground truth for them to be considered matching
        seen (torch.Tensor): How many images the network has already been trained on.
    """

    def __init__(self, coder, seen=0, coord_scale=1.0,
                 noobject_scale=1.0, object_scale=5.0, class_scale=1.0,
                 thresh=0.6, seen_thresh=12800):
        super(YoloLoss, self).__init__()
        coder = YoloCoder.coerce(coder)
        self.coder = coder
        self.classes = coder.classes
        self.num_classes = len(coder.classes)
        self.num_anchors = len(coder.anchors)
        self.anchor_step = len(coder.anchors[0])
        self.anchors = torch.Tensor(coder.anchors)
        self.register_buffer('seen', torch.tensor(seen))
        self.seen_thresh = seen_thresh

        self.coord_scale = coord_scale
        self.noobject_scale = noobject_scale
        self.object_scale = object_scale
        self.class_scale = class_scale
        self.thresh = thresh

        self.mse = nn.MSELoss(reduction='sum')
        self.clf = nn.CrossEntropyLoss(reduction='sum')

    @classmethod
    def demo(YoloLoss):
        coder = YoloCoder.demo()
        self = YoloLoss(coder)
        return self

    def demo_truth(self, bsize=2):
        """
        Example of truth variable
        """
        # true boxes for each item in the batch
        # each box encodes class, x_center, y_center, width, and height
        # coordinates are normalized in the range 0 to 1
        # items in each batch are padded with dummy boxes with class_id=-1
        cxywh = torch.FloatTensor([
            # boxes for batch item 1
            [[0.50, 0.50, 1.00, 1.00],
             [0.32, 0.42, 0.22, 0.12]],
            # boxes for batch item 2 (it has no objects, note the pad!)
            [[0, 0, 0, 0],
             [0, 0, 0, 0]],
        ])

        class_idxs = torch.FloatTensor([
            [2, 4],
            [-1, -1],
        ])

        weight = torch.FloatTensor([
            [0, 1],
            [-1, -1],
        ])

        if bsize == 0:
            cxywh = cxywh[0:0]
            weight = weight[0:0]
            class_idxs = class_idxs[0:0]

        if bsize != 2:
            cxywh = cxywh[0:1].repeat(bsize, 1, 1)
            weight = weight[0:1].repeat(bsize, 1)
            class_idxs = class_idxs[0:1].repeat(bsize, 1)

        target = {
            'cxywh': cxywh,
            'class_idxs': class_idxs,
            'weight': weight,
        }
        return target

    def forward(self, output, target, seen=None):
        """ Compute Region loss.

        CommandLine:
            xdoctest -m netharn.models.yolo2.yolo2 YoloLoss.forward

        Example:
            >>> # xdoc: +REQUIRES(--download, module:ndsampler)
            >>> from netharn.models.yolo2.yolo2 import *  # NOQA
            >>> model = Yolo2.demo()
            >>> self = YoloLoss(model.coder)
            >>> target = self.demo_truth()
            >>> output = model.coder.demo_output(bsize=target['cxywh'].shape[0])
            >>> print('target = ' + ub.repr2(ub.map_vals(lambda x: x.shape, target), nl=1))
            >>> print('output = ' + ub.repr2(ub.map_vals(lambda x: x.shape, output), nl=1))
            >>> seen = None
            >>> loss_parts = self(output, target)
            >>> print('loss_parts = ' + ub.repr2(loss_parts))

        Example:
            >>> # xdoc: +REQUIRES(--download, module:ndsampler)
            >>> # Test cases with no truth
            >>> from netharn.models.yolo2.yolo2 import *  # NOQA
            >>> self = YoloLoss.demo()
            >>> target = self.demo_truth(bsize=2)
            >>> output = self.coder.demo_output(bsize=2)
            >>> # Case 1: weights are -1
            >>> target['weight'][:] = -1
            >>> loss_parts = self(output, target)
            >>> # Case 2: empty with shape
            >>> target = {
            >>>     'cxywh': target['cxywh'][:, 0:0],
            >>>     'class_idxs': target['class_idxs'][:, 0:0],
            >>>     'weight': target['weight'][:, 0:0],
            >>> }
            >>> loss_parts = self(output, target)
            >>> # Case 3: empty without shape
            >>> target = {
            >>>     'cxywh': torch.Tensor(),
            >>>     'class_idxs': torch.Tensor(),
            >>>     'weight': torch.Tensor(),
            >>> }
            >>> loss_parts = self(output, target)

        Example:
            >>> # xdoc: +REQUIRES(--download, module:ndsampler)
            >>> from netharn.models.yolo2.yolo2 import *  # NOQA
            >>> info = dev_demodata()
            >>> self, output, target = ub.take(info, ['criterion', 'outputs', 'target'])
            >>> model = info['model']
            >>> self = YoloLoss(model.coder)
            >>> loss_parts = self.forward(output, target)
            >>> print('loss_parts = {!r}'.format(loss_parts))
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.figure(fnum=1, doclf=True)
            >>> sf = info['orig_sizes'][0]
            >>> dets.boxes.scale(sf, inplace=True)
            >>> kwplot.imshow(info['rgb255'], colorspace='rgb')
            >>> dets.draw()
            >>> kwplot.show_if_requested()
        """
        class_energy = output['class_energy']
        score_energy = output['score_energy']
        cxywh_energy = output['cxywh_energy']

        # Get x,y,w,h,conf,cls
        nB, nA, nC, nH, nW = class_energy.data.shape
        nC = self.num_classes
        assert nA == self.num_anchors
        assert nC == self.num_classes

        device = class_energy.device
        if seen is not None:
            self.seen = torch.tensor(seen)
        elif self.training:
            self.seen += nB

        if self.anchors.device != device:
            self.anchors = self.anchors.to(device)

        conf = score_energy.sigmoid()
        if nC > 1:
            # Swaps the dimensions from [B, A, C, H, W] to be [B, A, H, W, C]
            cls = class_energy.permute(0, 1, 3, 4, 2).contiguous()

        # Note: can't do inplace ops here.
        coord = torch.empty_like(cxywh_energy)
        coord[:, :, 0:2, :, :] = cxywh_energy[:, :, 0:2, :, :].sigmoid()    # cx,cy
        coord[:, :, 2:4, :, :] = cxywh_energy[:, :, 2:4, :, :]              # w,h

        info = self.coder.decode_batch(output, forloss=True)
        coord = info['coord']

        with torch.no_grad():
            # We only use the "decoded" coords (which are pred_boxes) to
            # determine the ground truth mask. We don't need to backprop
            # through here.

            # Get target values
            # CREATE ENCODED TRUTH VALUES *BASED ON THE OUTPUT*
            # TODO: refactor, minimize, verify correctness
            pred_boxes = info['pred_boxes']
            masks, truth = self.build_targets(pred_boxes, target, nB, nA, nC, nH, nW, device)

            if nC > 1:
                pcls_mask = masks['cls'].view(-1, 1)
                if _TORCH_HAS_BOOL_COMP:
                    tcls = truth['cls'][masks['cls'].bool()].view(-1).long()
                else:
                    tcls = truth['cls'][masks['cls']].view(-1).long()

        if nC > 1:
            if _TORCH_HAS_BOOL_COMP:
                pcls = cls.view(-1, nC)[pcls_mask.view(-1).bool()]
            else:
                pcls = cls.view(-1, nC)[pcls_mask.view(-1)]

        coord_mask = masks['coord'].repeat(1, 1, 4, 1)
        coord_ = coord.view(coord_mask.shape)
        conf_ = conf.view(masks['conf'].shape)

        # Compute losses
        coord_part = self.mse(coord_ * coord_mask, truth['coord'] * coord_mask)
        conf_part = self.mse(conf_ * masks['conf'], truth['conf'] * masks['conf'])

        loss_parts = {}
        loss_parts['coord'] = self.coord_scale * (coord_part / nB)
        loss_parts['conf'] = conf_part / nB
        if nC > 1 and pcls.numel() > 0:
            clf_part = (self.clf(pcls, tcls) / nB)
            loss_parts['cls'] = self.class_scale * 2 * clf_part

        return loss_parts

    def build_targets(self, pred_boxes, target, nB, nA, nC, nH, nW, device=None):
        """
        For each ground truth, assign it to a predicted box, and construct
        appropriate truth tensors and masks.

        Args:
            pred_boxes : OUTPUT_SPACE cywh boxes
            target : contains 01-NORMALIZED cxywh boxes, cidxs, and weights

        TODO:
            standardize nonrmalization of inputs (best case is probably to
            simply allow raw targets and normalize insidet this func)

        Example:
            >>> # xdoc: +REQUIRES(--download, module:ndsampler)
            >>> from netharn.models.yolo2.yolo2 import *  # NOQA
            >>> info = dev_demodata()
            >>> self, output, target = ub.take(info, ['criterion', 'outputs', 'target'])
            >>> decoded_info = self.coder.decode_batch(output, forloss=True)
            >>> pred_boxes = decoded_info['pred_boxes']
            >>> nB, nA, nC, nH, nW = output['class_energy'].shape
            >>> device = pred_boxes.device
            >>> self.seen += 100000
            >>> masks, truth = self.build_targets(pred_boxes, target, nB, nA, nC, nH, nW, device)
            >>> print('masks sum = {}'.format(ub.map_vals(lambda x: x.sum(), masks)))
            >>> print('truth sum = {}'.format(ub.map_vals(lambda x: x.sum(), truth)))
            >>> print('masks shape = {}'.format(ub.map_vals(lambda x: x.shape, masks)))
            >>> print('truth shape = {}'.format(ub.map_vals(lambda x: x.shape, truth)))

        Example:
            >>> # xdoc: +REQUIRES(--download, module:ndsampler)
            >>> # Test empty case
            >>> from netharn.models.yolo2.yolo2 import *  # NOQA
            >>> self = YoloLoss.demo()
            >>> target = self.demo_truth(bsize=2)
            >>> for k in target:
            >>>     target[k] = torch.Tensor()
            >>> output = self.coder.demo_output(bsize=2)
            >>> decoded_info = self.coder.decode_batch(output, forloss=True)
            >>> pred_boxes = decoded_info['pred_boxes']
            >>> nB, nA, nC, nH, nW = output['class_energy'].shape
            >>> device = pred_boxes.device
            >>> self.seen += 100000
            >>> masks, truth = self.build_targets(pred_boxes, target, nB, nA, nC, nH, nW, device)
            >>> print('masks sum = {}'.format(ub.map_vals(lambda x: x.sum(), masks)))
            >>> print('truth sum = {}'.format(ub.map_vals(lambda x: x.sum(), truth)))
            >>> print('masks shape = {}'.format(ub.map_vals(lambda x: x.shape, masks)))
            >>> print('truth shape = {}'.format(ub.map_vals(lambda x: x.shape, truth)))
            >>> print('masks sum = {}'.format(ub.map_vals(lambda x: x.sum(), masks)))
            >>> print('truth sum = {}'.format(ub.map_vals(lambda x: x.sum(), truth)))
            >>> print('masks shape = {}'.format(ub.map_vals(lambda x: x.shape, masks)))
            >>> print('truth shape = {}'.format(ub.map_vals(lambda x: x.shape, truth)))
        """
        import math
        # Parameters

        target_cxwh = target['cxywh']  # normalized
        target_cidx = target['class_idxs']
        target_weight = target['weight']

        # TODO: ensure normalized

        # nB = target_cxwh.size(0)
        # nA = self.num_anchors
        nPixels = nH * nW
        item_stride = nA * nH * nW

        # Tensors
        conf_mask = torch.full((nB, nA, nPixels), self.noobject_scale, requires_grad=False)
        coord_mask = torch.zeros(nB, nA, 1, nPixels, requires_grad=False)
        cls_mask = torch.zeros(nB, nA, nPixels, requires_grad=False).byte()

        tcoord = torch.zeros(nB, nA, 4, nPixels, requires_grad=False)
        tconf = torch.zeros(nB, nA, nPixels, requires_grad=False)
        tcls = torch.zeros(nB, nA, nPixels, requires_grad=False)

        # 12800
        if self.seen < self.seen_thresh:
            coord_mask.fill_(1)
            # coord_mask.fill_(.01 / self.coord_scale)

            if self.anchor_step == 4:
                # TODO: use permute
                tcoord[:, :, 0] = self.anchors[:, 2].contiguous().view(1, nA, 1, 1).repeat(nB, 1, 1, nPixels)
                tcoord[:, :, 1] = self.anchors[:, 3].contiguous().view(
                    1, nA, 1, 1).repeat(nB, 1, 1, nPixels)
            else:
                tcoord[:, :, 0].fill_(0.5)
                tcoord[:, :, 1].fill_(0.5)

        self.anchors = self.anchors.to(pred_boxes.device)
        if self.anchor_step == 4:
            anchors_wh = self.anchors[:, 2:4]
        else:
            anchors_wh = self.anchors

        if target_weight.numel() > 0:

            # Determine the number of true boxes in each batch item by checking
            # if the weight has a padding value (which should be -1)
            target_numgt = (target_weight >= 0).sum(dim=1).data.cpu().numpy()

            # For each batch item, ...
            for b, num_gt in enumerate(target_numgt):
                if num_gt > 0:
                    # Remove dummy batch padding (which should be trailing)
                    gtb = target_cxwh[b, 0:num_gt].view(-1, 4)
                    gtc = target_cidx[b, 0:num_gt].view(-1)
                    gtw = target_weight[b, 0:num_gt].view(-1)

                    # Build up tensors
                    cur_pred_boxes = pred_boxes[b * item_stride: (b + 1) * item_stride]

                    # convert to output space
                    cur_true_boxes = gtb.clone()
                    cur_true_boxes[:, ::2] *= nW
                    cur_true_boxes[:, 1::2] *= nH

                    # Set confidence mask of matching detections to 0, we
                    # will selectively reenable a subset of these values later
                    iou_gt_pred = bbox_ious(cur_true_boxes, cur_pred_boxes)
                    mask = (iou_gt_pred > self.thresh).sum(0) >= 1
                    conf_mask_b = conf_mask[b]
                    conf_mask_b[mask.view_as(conf_mask_b)] = 0

                    # Find best anchor for each gt
                    gt_wh = cur_true_boxes[:, 2:4]
                    iou_gt_anchors = wh_ious(gt_wh, anchors_wh)
                    _, best_anchors = iou_gt_anchors.max(1)

                    # Set masks and target values for each gt
                    for i in range(num_gt):
                        cx, cy, w, h = cur_true_boxes[i, 0:4]

                        gi = min(nW - 1, max(0, int(cx)))
                        gj = min(nH - 1, max(0, int(cy)))
                        anch_idx = best_anchors[i]
                        grid_idx = gj * nW + gi

                        weight = gtw[i]

                        coord_mask[b, anch_idx, 0, grid_idx] = (2 - (w * h) / nPixels) * weight
                        cls_mask[b, anch_idx, grid_idx] = weight
                        conf_mask[b, anch_idx, grid_idx] = self.object_scale * weight

                        tcoord[b, anch_idx, 0, grid_idx] = cx - gi
                        tcoord[b, anch_idx, 1, grid_idx] = cy - gj
                        tcoord[b, anch_idx, 2, grid_idx] = math.log(w / self.anchors[anch_idx, 0])
                        tcoord[b, anch_idx, 3, grid_idx] = math.log(h / self.anchors[anch_idx, 1])

                        iou = iou_gt_pred[i, anch_idx * nPixels + grid_idx]
                        tconf[b, anch_idx, grid_idx] = iou

                        tcls[b, anch_idx, grid_idx] = gtc[i]

        masks = {
            'coord': coord_mask.to(device).sqrt_(),
            'conf': conf_mask.to(device).sqrt_(),
            'cls': cls_mask.to(device),
        }
        truth = {
            'coord': tcoord.to(device),
            'conf': tconf.to(device),
            'cls': tcls.to(device),
        }
        return masks, truth


def wh_ious(wh1, wh2):
    """
    Compute IOU between centered boxes with given wh.
    Slightly faster than zeroing the center coords.
    """
    half_wh1 = (wh1 / 2)
    half_wh2 = (wh2 / 2)

    b1x2, b1y2 = (half_wh1).split(1, 1)
    b2x2, b2y2 = (half_wh2).split(1, 1)
    b1x1, b1y1 = -b1x2, -b1y2
    b2x1, b2y1 = -b2x2, -b2y2

    dx = (b1x2.min(b2x2.t()) - b1x1.max(b2x1.t())).clamp_(min=0)
    dy = (b1y2.min(b2y2.t()) - b1y1.max(b2y1.t())).clamp_(min=0)
    intersections = dx * dy

    areas1 = wh1.prod(dim=1, keepdim=True)
    areas2 = wh2.prod(dim=1, keepdim=True)

    unions = (areas1 + areas2.t()) - intersections
    return intersections / unions


def bbox_ious(boxes1, boxes2):
    """ Compute IOU between all boxes from ``boxes1`` with all boxes from ``boxes2``.

    Args:
        boxes1 (torch.Tensor): List of bounding boxes
        boxes2 (torch.Tensor): List of bounding boxes

    Note:
        List format: [[xc, yc, w, h],...]
    """
    b1x1, b1y1 = (boxes1[:, :2] - (boxes1[:, 2:4] / 2)).split(1, 1)
    b1x2, b1y2 = (boxes1[:, :2] + (boxes1[:, 2:4] / 2)).split(1, 1)
    b2x1, b2y1 = (boxes2[:, :2] - (boxes2[:, 2:4] / 2)).split(1, 1)
    b2x2, b2y2 = (boxes2[:, :2] + (boxes2[:, 2:4] / 2)).split(1, 1)

    dx = (b1x2.min(b2x2.t()) - b1x1.max(b2x1.t())).clamp(min=0)
    dy = (b1y2.min(b2y2.t()) - b1y1.max(b2y1.t())).clamp(min=0)
    intersections = dx * dy

    areas1 = (b1x2 - b1x1) * (b1y2 - b1y1)
    areas2 = (b2x2 - b2x1) * (b2y2 - b2y1)
    unions = (areas1 + areas2.t()) - intersections

    return intersections / unions


def find_anchors2(sampler, reduction=32, num_anchors=5):
    """
    Ignore:
        import ndsampler
        sampler = ndsampler.CocoSampler.demo('shapes')
        import sys, ubelt
        from netharn.models.yolo2.yolo2 import *  # NOQA
        sys.path.append(ubelt.expandpath('~/code/netharn/examples'))
        from object_detection import setup_harn
        cmdline = False
        kw = {
            'train_dataset': '~/data/VOC/voc-trainval.mscoco.json',
            'vali_dataset': '~/data/VOC/voc-test-2007.mscoco.json',
        }
        harn = setup_harn(**kw)
        torch_dset = harn.hyper.datasets['train']
        sampler = torch_dset.sampler
        anchors = find_anchors2(sampler, reduction)
        print('anchors = {!r}'.format(anchors))
    """
    from sklearn import cluster
    # all_cxywh = sampler.regions.targets._getcols(['cx', 'cy', 'width', 'height'])
    all_wh = sampler.regions.targets._getcols(['width', 'height'])

    all_imgwh = sampler.regions.targets._getcols(['img_width', 'img_height'])

    # foo = all_imgwh / reduction
    # foo * input_dims

    all_norm_wh = all_wh * reduction / all_imgwh

    ogrid_wh = 18
    all_ogrid_wh = all_norm_wh * ogrid_wh

    algo = cluster.KMeans(
        n_clusters=num_anchors, n_init=20, max_iter=10000, tol=1e-6,
        algorithm='elkan', verbose=0)
    algo.fit(all_ogrid_wh)
    anchors = algo.cluster_centers_
    return anchors


def find_anchors(dset):
    """
    OLD CODE, dont use

    Ignore:
        >>> # xdoctest: +SKIP
        >>> # xdoc: +REQUIRES(--download, module:ndsampler)
        >>> self = YoloVOCDataset(split='train', years=[2007])
        >>> anchors = self._find_anchors()
        >>> print('anchors = {}'.format(ub.repr2(anchors, precision=2)))
        >>> # xdoctest: +REQUIRES(--show)
        >>> xy = -anchors / 2
        >>> wh = anchors
        >>> show_boxes = np.hstack([xy, wh])
        >>> kwplot.autompl()  # xdoc: +SKIP
        >>> kwplot.figure(doclf=True, fnum=1)
        >>> kwplot.draw_boxes(show_boxes, box_format='tlwh')
        >>> from matplotlib import pyplot as plt
        >>> plt.gca().set_xlim(xy.min() - 1, wh.max() / 2 + 1)
        >>> plt.gca().set_ylim(xy.min() - 1, wh.max() / 2 + 1)
        >>> plt.gca().set_aspect('equal')
    """
    import numpy as np
    from PIL import Image
    from sklearn import cluster
    all_norm_wh = []
    for i in ub.ProgIter(range(len(dset)), desc='find anchors'):
        annots = dset._load_annotation(i)
        img_wh = np.array(Image.open(dset.gpaths[i]).size)
        boxes = np.array(annots['boxes'])
        box_wh = boxes[:, 2:4] - boxes[:, 0:2]
        # normalize to 0-1
        norm_wh = box_wh / img_wh
        all_norm_wh.extend(norm_wh.tolist())
    # Re-normalize to the size of the grid
    all_wh = np.array(all_norm_wh) * dset.base_wh[0] / dset.factor
    algo = cluster.KMeans(
        n_clusters=5, n_init=20, max_iter=10000, tol=1e-6,
        algorithm='elkan', verbose=0)
    algo.fit(all_wh)
    anchors = algo.cluster_centers_
    return anchors


def demo_voc_weights(key='lightnet'):
    """
    Demo weights for Pascal VOC dataset
    """

    if key == 'lightnet':
        # url = 'https://gitlab.com/EAVISE/lightnet/raw/master/examples/yolo-voc/lightnet_weights.pt'
        # hash_prefix = 'c4597fed8eb1b01da3495'
        fpath = ub.grabdata('https://data.kitware.com/api/v1/file/5c2e6e1a8d777f072bf2dc65/download',
                            fname='lightnet_weights.pt',
                            appname='netharn',
                            hasher='sha512',
                            hash_prefix='c4597fed8eb1b01')
        return fpath
    elif key == 'darknet':
        url = 'https://pjreddie.com/media/files/yolo-voc.weights'
        hash_prefix = '3033f5f510c25ab3ff6b9'
        fpath = ub.grabdata(url, appname='netharn', hash_prefix=hash_prefix)
        return fpath
    else:
        raise KeyError(key)
    # import lightnet
    # from os.path import dirname, join
    # dpath = dirname(dirname(lightnet.__file__))
    # fpath = join(dpath, 'examples', 'yolo-voc', 'lightnet_weights.pt')
    return fpath


def initial_imagenet_weights():
    # import os
    try:
        darknet_weight_fpath = ub.grabdata(
            'https://pjreddie.com/media/files/darknet19_448.conv.23',
            appname='netharn', hash_prefix='8016f5b7ddc15c5d7dad2315')
        torch_fpath = darknet_weight_fpath + '_lntf.pt'
        import os
        if not os.path.exists(torch_fpath):
            import lightnet.models
            # hack to transform initial state
            model = lightnet.models.Yolo2(classes=20)
            model.load_weights(darknet_weight_fpath)
            torch.save(model.state_dict(), torch_fpath)
    except (Exception, ImportError):
        # Maybe this had a weird bad init state?
        torch_fpath = ub.grabdata('https://data.kitware.com/api/v1/file/5b16b81c8d777f15ebe1ffce/download',
                                  fname='darknet19_448.conv.23.pt',
                                  appname='netharn',
                                  hasher='sha512',
                                  hash_prefix='f38968224a81a')
    return torch_fpath


def demo_image(inp_size):
    import kwimage
    import numpy as np
    import cv2
    rgb255 = kwimage.grab_test_image('astro', 'rgb')
    rgb01 = cv2.resize(rgb255, inp_size).astype(np.float32) / 255
    im_data = torch.FloatTensor([rgb01.transpose(2, 0, 1)])
    return im_data, rgb255


def dev_demodata():
    """
    Create demodata for each part of the system
    """
    inp_size = (288, 288)

    classes = ['aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus', 'car',
               'cat', 'chair', 'cow', 'diningtable', 'dog', 'horse',
               'motorbike', 'person', 'pottedplant', 'sheep', 'sofa', 'train',
               'tvmonitor']

    model = Yolo2(classes=classes, conf_thresh=0.01, nms_thresh=0.4).eval()
    state_dict = torch.load(demo_voc_weights())['weights']
    model.load_state_dict(state_dict)

    criterion = YoloLoss(model.coder)

    im_data, rgb255 = demo_image(inp_size)
    inputs = torch.cat([im_data, im_data])  # make a batch size of 2
    outputs = model(inputs)

    label = {
        'cxywh': torch.FloatTensor([
            [205, 255, 350, 500],
        ]) / 512,
        'class_idxs': torch.LongTensor([14]),
        'weight': torch.FloatTensor([1.0]),
    }
    import netharn as nh
    target = nh.data.collate.padded_collate([label, label])

    orig_sizes = torch.LongTensor([rgb255.shape[0:2][::-1]] * len(inputs))

    raw = torch.cat([
        outputs['cxywh_energy'],
        outputs['score_energy'],
        outputs['class_energy']], dim=2)

    demodata = {
        'model': model,
        'coder': model.coder,
        'inputs': inputs,
        'outputs': outputs,
        'criterion': criterion,
        'target': target,
        'rgb255': rgb255,
        'raw': raw,
        'orig_sizes': orig_sizes,
    }
    return demodata


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m netharn.models.yolo2.yolo2
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import glob
from os.path import expanduser, exists, join, basename
import numpy as np
import cv2
# try:
#     # Don't import skimage.io immediately because it imports pyplot
#     # See GH Issue https://github.com/scikit-image/scikit-image/issues/3347
#     import skimage.io
# except ImportError:
#     pass


def load_image_paths(dpath, ext=('.png', '.tiff', 'tif')):
    dpath = expanduser(dpath)
    if not exists(dpath):
        raise ValueError('dpath = {} does not exist'.format(dpath))
    if not isinstance(ext, (list, tuple)):
        ext = [ext]

    image_paths = []
    for ext_ in ext:
        image_paths.extend(list(glob.glob(join(dpath, '*' + ext_))))
    # potentially non-general
    # (utilfname solves this though)
    image_paths = sorted(image_paths, key=basename)
    return image_paths


def image_slices(img_shape, target_shape, overlap=0, keepbound=False):
    """
    Generates "sliding window" slices to break a large image into smaller
    pieces.

    Args:
        img_shape (tuple): height and width of the image

        target_shape (tuple): (height, width) of the

        overlap (float): a number between 0 and 1 indicating the fraction of
            overlap that parts will have. Must be `0 <= overlap < 1`.

        keepbound (bool): if True, a non-uniform step will be taken to ensure
            that the right / bottom of the image is returned as a slice if
            needed. Such a slice will not obey the overlap constraints.
            (Defaults to False)

    Yields:
        tuple(slice, slice): row and column slices used for numpy indexing

    Example:
        >>> img_shape = (2000, 2000)
        >>> target_shape = (360, 480)
        >>> overlap = 0
        >>> keepbound = True
        >>> list(image_slices(img_shape, target_shape, overlap, keepbound))
    """
    if overlap < 0 or overlap >= 1:
        raise ValueError(('part overlap was {}, but it must be '
                          'in the range [0, 1)').format(overlap))
    ph, pw = target_shape
    sy = int(ph - ph * overlap)
    sx = int(pw - pw * overlap)
    orig_h, orig_w = img_shape
    kw = dict(keepbound=keepbound, check=False, start=0)
    for rslice in wide_strides_1d(ph, orig_h, sy, **kw):
        for cslice in wide_strides_1d(pw, orig_w, sx, **kw):
            yield rslice, cslice


def overlay_colorized(colorized, orig, alpha=.6, keepcolors=False):
    """
    Overlays a color segmentation mask on an original image

    Args:
        colorized (ndarray): the color mask to be overlayed on top of the original image
        orig (ndarray): the original image to superimpose on
        alpha (float): blend level to use if colorized is not an alpha image

    """
    import kwimage
    color_mask = kwimage.ensure_alpha_channel(colorized, alpha=alpha)
    if not keepcolors:
        orig = ensure_grayscale(orig)
    color_blend = kwimage.overlay_alpha_images(color_mask, orig)
    color_blend = (color_blend * 255).astype(np.uint8)
    return color_blend


def wide_strides_1d(margin, stop, step=None, start=0, keepbound=False,
                    check=True):
    """
    Helper for `image_slices`. Generates slices in a single dimension.

    Args:
        start (int): starting point (in most cases set this to 0)

        margin (int): the length of the slice (window)

        stop (int): the length of the image dimension

        step (int): the length of each step / distance between slices

        keepbound (bool): if True, a non-uniform step will be taken to ensure
            that the right / bottom of the image is returned as a slice if
            needed. Such a slice will not obey the overlap constraints.
            (Defaults to False)

        check (bool): if True an error will be raised if the window does not
            cover the entire extent from start to stop, even if keepbound is
            True.

    Yields:
        slice : slice in one dimension of size (margin)

    Example:
        >>> stop, margin, step = 2000, 360, 360
        >>> keepbound = True
        >>> strides = list(wide_strides_1d(margin, stop, step, keepbound, check=False))
        >>> assert all([(s.stop - s.start) == margin for s in strides])

    Example:
        >>> stop, margin, step = 200, 46, 7
        >>> keepbound = True
        >>> strides = list(wide_strides_1d(margin, stop, step, keepbound=False, check=True))
        >>> starts = np.array([s.start for s in strides])
        >>> stops = np.array([s.stop for s in strides])
        >>> widths = stops - starts
        >>> assert np.all(np.diff(starts) == step)
        >>> assert np.all(widths == margin)

    Example:
        >>> import pytest
        >>> stop, margin, step = 200, 36, 7
        >>> with pytest.raises(ValueError):
        ...     list(wide_strides_1d(margin, stop, step))
    """
    if step is None:
        step = margin

    if check:
        # see how far off the end we would fall if we didnt check bounds
        perfect_final_pos = (stop - start - margin)
        overshoot = perfect_final_pos % step
        if overshoot > 0:
            raise ValueError(
                ('margin={} and step={} overshoot endpoint={} '
                 'by {} units when starting from={}').format(
                     margin, step, stop, overshoot, start))
    pos = start
    # probably could be more efficient with numpy here
    while True:
        endpos = pos + margin
        yield slice(pos, endpos)
        # Stop once we reached the end
        if endpos == stop:
            break
        pos += step
        if pos + margin > stop:
            if keepbound:
                # Ensure the boundary is always used even if steps
                # would overshoot Could do some other strategy here
                pos = stop - margin
            else:
                break


def get_num_channels(img):
    """ Returns the number of color channels """
    ndims = len(img.shape)
    if ndims == 2:
        nChannels = 1
    elif ndims == 3 and img.shape[2] == 3:
        nChannels = 3
    elif ndims == 3 and img.shape[2] == 4:
        nChannels = 4
    elif ndims == 3 and img.shape[2] == 1:
        nChannels = 1
    else:
        raise ValueError('Cannot determine number of channels '
                         'for img.shape={}'.format(img.shape))
    return nChannels


def ensure_grayscale(img, colorspace_hint='BGR'):
    """
    Example:
        >>> import numpy as np
        >>> ensure_grayscale(np.array([[[0, 0, 1]]], dtype=np.float32))
        array([[0.299]], dtype=float32)
    """
    import kwimage
    img = kwimage.ensure_float01(img, copy=False)
    c = get_num_channels(img)
    if c == 1:
        return img
    else:
        return kwimage.convert_colorspace(img, src_space=colorspace_hint,
                                          dst_space='gray')


def adjust_gamma(img, gamma=1.0):
    """
    gamma correction function

    References:
        http://www.pyimagesearch.com/2015/10/05/opencv-gamma-correction/

    Ignore:
        >>> from netharn import util
        >>> fpath = util.grab_test_image()
        >>> img = util.imread(fpath)
        >>> gamma = .5
        >>> imgf = ensure_float01(img)
        >>> img2 = adjust_gamma(img, gamma)
        >>> img3 = adjust_gamma(imgf, gamma)
        >>> import plottool as pt
        >>> pt.qtensure()
        >>> pt.imshow(img, pnum=(3, 3, 1), fnum=1)
        >>> pt.imshow(img2, pnum=(3, 3, 2), fnum=1)
        >>> pt.imshow(img3, pnum=(3, 3, 3), fnum=1)
        >>> pt.imshow(adjust_gamma(img, 1), pnum=(3, 3, 5), fnum=1)
        >>> pt.imshow(adjust_gamma(imgf, 1), pnum=(3, 3, 6), fnum=1)
        >>> pt.imshow(adjust_gamma(img, 2), pnum=(3, 3, 8), fnum=1)
        >>> pt.imshow(adjust_gamma(imgf, 2), pnum=(3, 3, 9), fnum=1)
    """
    if img.dtype.kind in ('i', 'u'):
        # build a lookup table mapping the pixel values [0, 255] to
        # their adjusted gamma values
        table = (((np.arange(0, 256) / 255.0) ** (1 / gamma)) * 255).astype(np.uint8)
        invGamma = 1.0 / gamma
        table = np.array([
            ((i / 255.0) ** invGamma) * 255 for i in np.arange(0, 256)
        ]).astype("uint8")
        # apply gamma correction using the lookup table
        return cv2.LUT(img, table)
    else:
        import kwimage
        np_img = kwimage.ensure_float01(img, copy=False)
        gain = 1
        np_img = gain * (np_img ** (1 / gamma))
        np_img = np.clip(np_img, 0, 1)
        return np_img


def _lookup_cv2_colorspace_conversion_code(src_space, dst_space):
    src = src_space.upper()
    dst = dst_space.upper()
    convert_attr = 'COLOR_{}2{}'.format(src, dst)
    if not hasattr(cv2, convert_attr):
        prefix = 'COLOR_{}2'.format(src)
        valid_dst_spaces = [
            key.replace(prefix, '')
            for key in cv2.__dict__.keys() if key.startswith(prefix)]
        raise KeyError(
            '{} does not exist, valid conversions from {} are to {}'.format(
                convert_attr, src_space, valid_dst_spaces))
    else:
        code = getattr(cv2, convert_attr)
    return code


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m netharn.util.imutil all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
